
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "defines.h"
#include "collision.h"
#include "oload.h"
#include "txt.h"

/*
#define COLLISION_DEBUG 1
*/

extern struct object_object **world_map;
extern struct game_object **game_object_map;
extern int world_x, world_y, world_z;
struct vector gravity;
double collision_txt_speed;
int collision_occurred, collision_txtid;
#ifdef COLLISION_DEBUG
int tested_faces, tested_long, tested_evil, tested_passes;
#endif


struct object_object *colliders_o;
struct game_object *colliders_g;
struct vector new_velocity_vector;


int collision_init(void) {

  gravity.x = 0;
  gravity.y = -0.8;
  gravity.z = 0;

  return SUCCEEDED;
}


void collision_closest_point_on_triangle(struct vector *a, struct vector *b, struct vector *c, struct vector *p, struct vector *r) {

  struct vector rab, rbc, rca;
  double r1, r2, r3, x, y, z;


  collision_closest_point_on_line(a, b, p, &rab);
  collision_closest_point_on_line(b, c, p, &rbc);
  collision_closest_point_on_line(c, a, p, &rca);
  x = rab.x - p->x;
  y = rab.y - p->y;
  z = rab.z - p->z;
  r1 = x*x + y*y + z*z;
  x = rbc.x - p->x;
  y = rbc.y - p->y;
  z = rbc.z - p->z;
  r2 = x*x + y*y + z*z;
  x = rca.x - p->x;
  y = rca.y - p->y;
  z = rca.z - p->z;
  r3 = x*x + y*y + z*z;
  if (r1 <= r2 && r1 <= r3) {
    r->x = rab.x;
    r->y = rab.y;
    r->z = rab.z;
    return;
  }
  if (r2 < r1 && r2 < r3) {
    r->x = rbc.x;
    r->y = rbc.y;
    r->z = rbc.z;
    return;
  }
  r->x = rca.x;
  r->y = rca.y;
  r->z = rca.z;
  return;
}


void collision_closest_point_on_line(struct vector *a, struct vector *b, struct vector *p, struct vector *r) {

  struct vector c, v;
  float d, t;


  /* determine t (the length of the vector from 'a' to 'p') */
  c.x = p->x - a->x;
  c.y = p->y - a->y;
  c.z = p->z - a->z;
  v.x = b->x - a->x;
  v.y = b->y - a->y;
  v.z = b->z - a->z;
  d = v.x*v.x + v.y*v.y + v.z*v.z;
  d = sqrt(d);
  v.x /= d;
  v.y /= d;
  v.z /= d;
  t = v.x*c.x + v.y*c.y + v.z*c.z;

  /* check to see if 't' is beyond the extents of the line segment */
  if (t < 0.0) {
    r->x = a->x;
    r->y = a->y;
    r->z = a->z;
    return;
  }
  if (t > d) {
    r->x = b->x;
    r->y = b->y;
    r->z = b->z;
    return;
  }

  /* return the point between 'a' and 'b' */
  r->x = a->x + (v.x * t);
  r->y = a->y + (v.y * t);
  r->z = a->z + (v.z * t);
  return;
}


int is_point_inside_a_triangle(struct vector *a, struct vector *b, struct vector *c, struct vector *p) {

  struct vector an, bn, cn;
  float l, d;


  an.x = a->x - p->x;
  an.y = a->y - p->y;
  an.z = a->z - p->z;
  bn.x = b->x - p->x;
  bn.y = b->y - p->y;
  bn.z = b->z - p->z;
  cn.x = c->x - p->x;
  cn.y = c->y - p->y;
  cn.z = c->z - p->z;
  l = an.x*an.x + an.y*an.y + an.z*an.z;
  l = sqrt(l);
  an.x /= l;
  an.y /= l;
  an.z /= l;
  d = bn.x*bn.x + bn.y*bn.y + bn.z*bn.z;
  d = sqrt(d);
  bn.x /= d;
  bn.y /= d;
  bn.z /= d;
  l = cn.x*cn.x + cn.y*cn.y + cn.z*cn.z;
  l = sqrt(l);
  cn.x /= l;
  cn.y /= l;
  cn.z /= l;

  d = an.x*bn.x + an.y*bn.y + an.z*bn.z;
  if (d > 1)
    l = 0;     /* l = acos(1); */
  else if (d < -1)
    l = M_PI;  /* l = acos(-1); */
  else
    l = acos(d);

  d = bn.x*cn.x + bn.y*cn.y + bn.z*cn.z;
  if (d > 1)
    ;          /* l += acos(1); */
  else if (d < -1)
    l += M_PI; /* l += acos(-1); */
  else
    l += acos(d);

  d = an.x*cn.x + an.y*cn.y + an.z*cn.z;
  if (d > 1)
    ;          /* l += acos(1); */
  else if (d < -1)
    l += M_PI; /* l += acos(-1); */
  else
    l += acos(d);

  if (l >= (M_PI * 2) - EPSILON)
    return SUCCEEDED;

  return FAILED;
}


double collision_intersect_sphere(struct vector *ro, struct vector *rv, struct vector *so) {

  float c, v, d, x, y, z;


  /* note that rv is normalized, and rv will be inverted! */
  x = so->x - ro->x;
  y = so->y - ro->y;
  z = so->z - ro->z;
  c = x*x + y*y + z*z;
  c = sqrt(c);
  v = x*rv->x + y*rv->y + z*rv->z;
  d = 1.0 - (c*c - v*v);

  /* if there was no intersection, return */
  if (d < 0)
    return d;

  /* return the distance to the first intersecting point */
  return -v - sqrt(d);

}


double collision_intersect(struct vector *p_origin, struct vector *p_normal, struct vector *r_origin, struct vector *r_vector) {

  float a, n, d;


  d = p_normal->x * r_vector->x + p_normal->y * r_vector->y + p_normal->z * r_vector->z;
  if (d == 0)
    return -1;
  a = p_normal->x * p_origin->x + p_normal->y * p_origin->y + p_normal->z * p_origin->z;
  n = p_normal->x * r_origin->x + p_normal->y * r_origin->y + p_normal->z * r_origin->z - a;

  return -(n / d);
}


int collision_detection(struct vector *source_point, struct vector *velocity, struct vector *ellipsoid) {

  /* scale */
  source_point->x /= ellipsoid->x;
  source_point->y /= ellipsoid->y;
  source_point->z /= ellipsoid->z;
  velocity->x /= ellipsoid->x;
  velocity->y /= ellipsoid->y;
  velocity->z /= ellipsoid->z;

  /* collision status variables */
  collision_occurred = NO;

  new_velocity_vector.x = 0;
  new_velocity_vector.y = 0;
  new_velocity_vector.z = 0;

#ifdef COLLISION_DEBUG
  tested_faces = 0;
  tested_long = 0;
  tested_evil = 0;
  tested_passes = 0;
#endif

  collision_with_world(source_point, velocity, ellipsoid);

  new_velocity_vector.x *= ellipsoid->x;
  new_velocity_vector.y *= ellipsoid->y;
  new_velocity_vector.z *= ellipsoid->z;

#ifdef COLLISION_DEBUG
  fprintf(stderr, "passes: %d, total listed faces: %d, cumbersome: %d, all the way: %d\n", tested_passes, tested_faces, tested_long, tested_evil);
#endif

  /* back scale */
  source_point->x *= ellipsoid->x;
  source_point->y *= ellipsoid->y;
  source_point->z *= ellipsoid->z;

  return SUCCEEDED;
}


int collision_with_world(struct vector *source_point, struct vector *velocity, struct vector *ellipsoid) {

  /*
  struct vector nearest_intersection_point, intersection_point;
  */
  struct vector unscaled_source_point, ellipsoid_inv;
  struct vector polygon_intersection_point, nearest_polygon_intersection_point;
  struct vector slide_plane_origin, slide_plane_vector, slide_plane_normal;
  struct vector plane_intersection_point, destination_point;
  struct vector new_destination_point;
  struct vector p_origin, p_normal, velocity_n, p_2, p_3;
  struct object_object *w;
  struct game_object *g;
  struct mesh2 *m;
  struct txt_object *txto;
  float j, k, l, nearest_distance = -1.0, distance_to_travel, ball_rs;
  float *tr, *fn, *vl;
  int collision_found = NO, i, tx, txt, q, z, *od;


  /* very small distance - no need to check collisions */
  distance_to_travel = velocity->x*velocity->x + velocity->y*velocity->y + velocity->z*velocity->z;
  if (distance_to_travel < EPSILON)
    return SUCCEEDED;
  distance_to_travel = sqrt(distance_to_travel);

#ifdef COLLISION_DEBUG
  tested_passes++;
#endif

  /* compute inverted radius vector */
  ellipsoid_inv.x = 1/ellipsoid->x;
  ellipsoid_inv.y = 1/ellipsoid->y;
  ellipsoid_inv.z = 1/ellipsoid->z;

  /* prepare for distance checks */
  unscaled_source_point.x = source_point->x * ellipsoid->x;
  unscaled_source_point.y = source_point->y * ellipsoid->y;
  unscaled_source_point.z = source_point->z * ellipsoid->z;

  /* the squared bounding sphere's radius for movement volume */
  j = (fabs(velocity->x)+1)*ellipsoid->x;
  k = (fabs(velocity->y)+1)*ellipsoid->y;
  l = (fabs(velocity->z)+1)*ellipsoid->z;
  if (j > k && j > l)
    ball_rs = j;
  else if (k > j && k > l)
    ball_rs = k;
  else
    ball_rs = l;
  if (ball_rs < ellipsoid->x)
    ball_rs = ellipsoid->x;
  if (ball_rs < ellipsoid->y)
    ball_rs = ellipsoid->y;
  if (ball_rs < ellipsoid->z)
    ball_rs = ellipsoid->z;

  /* compute normalized velocity */
  l = sqrt(velocity->x*velocity->x + velocity->y*velocity->y + velocity->z*velocity->z);
  velocity_n.x = velocity->x / l;
  velocity_n.y = velocity->y / l;
  velocity_n.z = velocity->z / l;

  /* collect all potential colliders */
  collision_collect_colliders(&unscaled_source_point);

  txt = -1;

  /* test collisions against the world */
  w = colliders_o;
  while (w != NULL) {
    m = w->mesh;
    if (m->collision_status == OFF) {
      w = w->next_ext;
      continue;
    }

    /* do the bounding sphere test */
    j = w->sdi.bounding_sphere[BS_X] - unscaled_source_point.x;
    k = w->sdi.bounding_sphere[BS_Y] - unscaled_source_point.y;
    l = w->sdi.bounding_sphere[BS_Z] - unscaled_source_point.z;
    l = sqrt(l*l + k*k + j*j);
    if (l > w->sdi.bounding_sphere[BS_R] + ball_rs) {
      w = w->next_ext;
      continue;
    }

    tr = m->triangle_radius_list;
    fn = m->face_normal_list;
    vl = m->vertex_list;
    od = m->object_data;

    /* test the objects one by one */
    for (q = 0; q < m->objects; q++) {
      tx = *(od++);
      z = *(od++);

      /* does the texture allow collision? */
      if (txt != tx) {
	txt = tx;
	txt_get(txt, &txto);
      }
      if (txto->status & TXT_STATUS_NO_COLLISION) {
	tr += z;
	fn += z*3;
	vl += z*3*3;
	continue;
      }

      for (i = 0; i < z; i++) {
#ifdef COLLISION_DEBUG
	tested_faces++;
#endif
	/* load vertice */
	p_origin.x = *(vl++);
	p_origin.y = *(vl++);
	p_origin.z = *(vl++);

	/* compute distance to face ball */
	j = p_origin.x - unscaled_source_point.x;
	k = p_origin.y - unscaled_source_point.y;
	l = p_origin.z - unscaled_source_point.z;
	l = sqrt(l*l + k*k + j*j);
	if (l > *tr + ball_rs) {
	  tr++;
	  fn += 3;
	  vl += 3*2;
	  continue;
	}

	p_origin.x *= ellipsoid_inv.x;
	p_origin.y *= ellipsoid_inv.y;
	p_origin.z *= ellipsoid_inv.z;

	/* scale the plane normal */
	p_normal.x = *(fn++) * ellipsoid->x;
	p_normal.y = *(fn++) * ellipsoid->y;
	p_normal.z = *(fn++) * ellipsoid->z;

	l = p_normal.x*p_normal.x + p_normal.y*p_normal.y + p_normal.z*p_normal.z;
	l = sqrt(l);
	p_normal.x /= l;
	p_normal.y /= l;
	p_normal.z /= l;

	/* determine the distance from the plane to the source */
	l = (source_point->x - p_origin.x) * p_normal.x + (source_point->y - p_origin.y) * p_normal.y + (source_point->z - p_origin.z) * p_normal.z;

	if (l < 0) {
	  vl += 3*2;
	  continue;
	}

#ifdef COLLISION_DEBUG
	tested_long++;
#endif

	/* is the plane embedded? */
	if (l <= 1.0) {
	  plane_intersection_point.x = source_point->x - (p_normal.x * l);
	  plane_intersection_point.y = source_point->y - (p_normal.y * l);
	  plane_intersection_point.z = source_point->z - (p_normal.z * l);
	}
	else {
	  /* calculate the sphere intersection point */
	  plane_intersection_point.x = source_point->x - p_normal.x;
	  plane_intersection_point.y = source_point->y - p_normal.y;
	  plane_intersection_point.z = source_point->z - p_normal.z;
	  /* calculate the plane intersection point */
	  l = collision_intersect(&p_origin, &p_normal, &plane_intersection_point, &velocity_n);
	  /* are we travelling away from this polygon */
	  if (l < 0.0) {
	    vl += 3*2;
	    continue;
	  }
	  /* calculate the plane intersection point */
	  plane_intersection_point.x += velocity_n.x * l;
	  plane_intersection_point.y += velocity_n.y * l;
	  plane_intersection_point.z += velocity_n.z * l;
	}

#ifdef COLLISION_DEBUG
	tested_evil++;
#endif

	/* load the rest vertices */
	p_2.x = *(vl++) * ellipsoid_inv.x;
	p_2.y = *(vl++) * ellipsoid_inv.y;
	p_2.z = *(vl++) * ellipsoid_inv.z;
	p_3.x = *(vl++) * ellipsoid_inv.x;
	p_3.y = *(vl++) * ellipsoid_inv.y;
	p_3.z = *(vl++) * ellipsoid_inv.z;

	/* is the plane intersection point inside the triangle? */
	if (is_point_inside_a_triangle(&p_origin, &p_2, &p_3, &plane_intersection_point) == FAILED)
	  collision_closest_point_on_triangle(&p_origin, &p_2, &p_3, &plane_intersection_point, &polygon_intersection_point);
	else {
	  polygon_intersection_point.x = plane_intersection_point.x;
	  polygon_intersection_point.y = plane_intersection_point.y;
	  polygon_intersection_point.z = plane_intersection_point.z;
	}

	/* reverse intersect with the sphere */
	l = collision_intersect_sphere(&polygon_intersection_point, &velocity_n, source_point);

	/* was there an intersection with the sphere? */
	if (l >= 0.0 && l <= distance_to_travel) {
	  /*
	  intersection_point.x = polygon_intersection_point.x - (velocity_n.x * l);
	  intersection_point.y = polygon_intersection_point.y - (velocity_n.y * l);
	  intersection_point.z = polygon_intersection_point.z - (velocity_n.z * l);
	  */

	  /* closest intersection so far? */
	  if (collision_found == NO || l < nearest_distance) {
	    nearest_distance = l;
	    /*
	    nearest_intersection_point.x = intersection_point.x;
	    nearest_intersection_point.y = intersection_point.y;
	    nearest_intersection_point.z = intersection_point.z;
	    */
	    nearest_polygon_intersection_point.x = polygon_intersection_point.x;
	    nearest_polygon_intersection_point.y = polygon_intersection_point.y;
	    nearest_polygon_intersection_point.z = polygon_intersection_point.z;

	    /* save for further analyzis */
	    slide_plane_normal.x = p_normal.x;
	    slide_plane_normal.y = p_normal.y;
	    slide_plane_normal.z = p_normal.z;

	    /* save texture */
	    collision_txtid = tx;
	    collision_found = YES;
	  }
	}
      }
    }
    w = w->next_ext;
  }

  /* test collisions against the game objects */
  g = colliders_g;
  while (g != NULL) {
    object_get(g->object, 1, &w);
    vl = g->rotated_vertices;
    if (vl == NULL) {
      g = g->next_ext;
      continue;
    }

    /* do the bounding sphere test */
    j = g->x + w->sdi.bounding_sphere[BS_X] - unscaled_source_point.x;
    k = g->y + w->sdi.bounding_sphere[BS_Y] - unscaled_source_point.y;
    l = g->z + w->sdi.bounding_sphere[BS_Z] - unscaled_source_point.z;
    l = sqrt(j*j + k*k + l*l);
    if (l > w->sdi.bounding_sphere[BS_R] + ball_rs) {
      g = g->next_ext;
      continue;
    }

    m = w->mesh;
    tr = m->triangle_radius_list;
    fn = m->face_normal_list;
    od = m->object_data;

    /* test the objects one by one */
    for (q = 0; q < m->objects; q++) {
      tx = *(od++);
      z = *(od++);

      /* does the texture allow collision? */
      if (txt != tx) {
	txt = tx;
	txt_get(txt, &txto);
      }
      if (txto->status & TXT_STATUS_NO_COLLISION) {
	tr += z;
	fn += z*3;
	vl += z*3*3;
	continue;
      }

      for (i = 0; i < z; i++) {
#ifdef COLLISION_DEBUG
	tested_faces++;
#endif
	/* load vertice */
	p_origin.x = *(vl++);
	p_origin.y = *(vl++);
	p_origin.z = *(vl++);

	/* compute distance to face ball */
	j = p_origin.x - unscaled_source_point.x;
	k = p_origin.y - unscaled_source_point.y;
	l = p_origin.z - unscaled_source_point.z;
	l = sqrt(l*l + k*k + j*j);
	if (l > *tr + ball_rs) {
	  tr++;
	  fn += 3;
	  vl += 3*2;
	  continue;
	}

	p_origin.x *= ellipsoid_inv.x;
	p_origin.y *= ellipsoid_inv.y;
	p_origin.z *= ellipsoid_inv.z;

	/* scale the plane normal */
	p_normal.x = *(fn++) * ellipsoid->x;
	p_normal.y = *(fn++) * ellipsoid->y;
	p_normal.z = *(fn++) * ellipsoid->z;

	l = p_normal.x*p_normal.x + p_normal.y*p_normal.y + p_normal.z*p_normal.z;
	l = sqrt(l);
	p_normal.x /= l;
	p_normal.y /= l;
	p_normal.z /= l;

	/* determine the distance from the plane to the source */
	l = (source_point->x - p_origin.x) * p_normal.x + (source_point->y - p_origin.y) * p_normal.y + (source_point->z - p_origin.z) * p_normal.z;

	if (l < 0) {
	  vl += 3*2;
	  continue;
	}

#ifdef COLLISION_DEBUG
	tested_long++;
#endif

	/* is the plane embedded? */
	if (l <= 1.0) {
	  plane_intersection_point.x = source_point->x - (p_normal.x * l);
	  plane_intersection_point.y = source_point->y - (p_normal.y * l);
	  plane_intersection_point.z = source_point->z - (p_normal.z * l);
	}
	else {
	  /* calculate the sphere intersection point */
	  plane_intersection_point.x = source_point->x - p_normal.x;
	  plane_intersection_point.y = source_point->y - p_normal.y;
	  plane_intersection_point.z = source_point->z - p_normal.z;
	  /* calculate the plane intersection point */
	  l = collision_intersect(&p_origin, &p_normal, &plane_intersection_point, &velocity_n);
	  /* are we travelling away from this polygon */
	  if (l < 0.0) {
	    vl += 3*2;
	    continue;
	  }
	  /* calculate the plane intersection point */
	  plane_intersection_point.x += velocity_n.x * l;
	  plane_intersection_point.y += velocity_n.y * l;
	  plane_intersection_point.z += velocity_n.z * l;
	}

#ifdef COLLISION_DEBUG
	tested_evil++;
#endif

	/* load the rest vertices */
	p_2.x = *(vl++) * ellipsoid_inv.x;
	p_2.y = *(vl++) * ellipsoid_inv.y;
	p_2.z = *(vl++) * ellipsoid_inv.z;
	p_3.x = *(vl++) * ellipsoid_inv.x;
	p_3.y = *(vl++) * ellipsoid_inv.y;
	p_3.z = *(vl++) * ellipsoid_inv.z;

	/* is the plane intersection point inside the triangle? */
	if (is_point_inside_a_triangle(&p_origin, &p_2, &p_3, &plane_intersection_point) == FAILED)
	  collision_closest_point_on_triangle(&p_origin, &p_2, &p_3, &plane_intersection_point, &polygon_intersection_point);
	else {
	  polygon_intersection_point.x = plane_intersection_point.x;
	  polygon_intersection_point.y = plane_intersection_point.y;
	  polygon_intersection_point.z = plane_intersection_point.z;
	}

	/* reverse intersect with the sphere */
	l = collision_intersect_sphere(&polygon_intersection_point, &velocity_n, source_point);

	/* was there an intersection with the sphere? */
	if (l >= 0.0 && l <= distance_to_travel) {
	  /*
	  intersection_point.x = polygon_intersection_point.x - (velocity_n.x * l);
	  intersection_point.y = polygon_intersection_point.y - (velocity_n.y * l);
	  intersection_point.z = polygon_intersection_point.z - (velocity_n.z * l);
	  */

	  /* closest intersection so far? */
	  if (collision_found == NO || l < nearest_distance) {
	    nearest_distance = l;
	    /*
	    nearest_intersection_point.x = intersection_point.x;
	    nearest_intersection_point.y = intersection_point.y;
	    nearest_intersection_point.z = intersection_point.z;
	    */
	    nearest_polygon_intersection_point.x = polygon_intersection_point.x;
	    nearest_polygon_intersection_point.y = polygon_intersection_point.y;
	    nearest_polygon_intersection_point.z = polygon_intersection_point.z;

	    /* save for further analyzis */
	    slide_plane_normal.x = p_normal.x;
	    slide_plane_normal.y = p_normal.y;
	    slide_plane_normal.z = p_normal.z;

	    /* save texture */
	    collision_txtid = tx;
	    collision_found = YES;
	  }
	}
      }
    }
    g = g->next_ext;
  }

  /* no collisions? */
  if (collision_found == NO) {
    source_point->x += velocity->x;
    source_point->y += velocity->y;
    source_point->z += velocity->z;
    return SUCCEEDED;
  }

  txt_get(collision_txtid, &txto);
  collision_txtid = txto->id;
  collision_txt_speed = txto->speed;

  collision_occurred = YES;

  /* move to the nearest collision */
  source_point->x += velocity_n.x * (nearest_distance - EPSILON);
  source_point->y += velocity_n.y * (nearest_distance - EPSILON);
  source_point->z += velocity_n.z * (nearest_distance - EPSILON);

  /* what's our destination relative to the point of contact? */
  destination_point.x = nearest_polygon_intersection_point.x + velocity_n.x * (distance_to_travel - nearest_distance);
  destination_point.y = nearest_polygon_intersection_point.y + velocity_n.y * (distance_to_travel - nearest_distance);
  destination_point.z = nearest_polygon_intersection_point.z + velocity_n.z * (distance_to_travel - nearest_distance);

  /* determine the sliding plane */
  slide_plane_origin.x = nearest_polygon_intersection_point.x;
  slide_plane_origin.y = nearest_polygon_intersection_point.y;
  slide_plane_origin.z = nearest_polygon_intersection_point.z;
  slide_plane_normal.x = nearest_polygon_intersection_point.x - source_point->x;
  slide_plane_normal.y = nearest_polygon_intersection_point.y - source_point->y;
  slide_plane_normal.z = nearest_polygon_intersection_point.z - source_point->z;
  l = slide_plane_normal.x * slide_plane_normal.x + slide_plane_normal.y * slide_plane_normal.y + slide_plane_normal.z * slide_plane_normal.z;
  l = -sqrt(l);
  slide_plane_normal.x /= l;
  slide_plane_normal.y /= l;
  slide_plane_normal.z /= l;

  /* project the destination point onto the sliding plane */
  l = collision_intersect(&slide_plane_origin, &slide_plane_normal, &destination_point, &slide_plane_normal);

  /* save slide plane vector for later use */
  slide_plane_vector.x = slide_plane_normal.x * l;
  slide_plane_vector.y = slide_plane_normal.y * l;
  slide_plane_vector.z = slide_plane_normal.z * l;
  new_destination_point.x = destination_point.x + slide_plane_vector.x;
  new_destination_point.y = destination_point.y + slide_plane_vector.y;
  new_destination_point.z = destination_point.z + slide_plane_vector.z;

  /* generate the slide vector, which will become our new velocity vector for the next iteration */
  new_velocity_vector.x = new_destination_point.x - nearest_polygon_intersection_point.x;
  new_velocity_vector.y = new_destination_point.y - nearest_polygon_intersection_point.y;
  new_velocity_vector.z = new_destination_point.z - nearest_polygon_intersection_point.z;

  /* introduce friction */
  new_velocity_vector.x *= txto->friction;
  new_velocity_vector.y *= txto->friction;
  new_velocity_vector.z *= txto->friction;

  /* recursively slide */
  collision_with_world(source_point, &new_velocity_vector, ellipsoid);

  return SUCCEEDED;
}


void collision_collect_colliders(struct vector *source_point) {

  struct object_object *o, *p = NULL, **w;
  struct game_object *g, **h;
  int x, y, z, a, b, c, d, e, f;


  colliders_o = NULL;
  colliders_g = NULL;

  x = (source_point->x / CUBE_SIDE) - 1;
  y = (source_point->y / CUBE_SIDE) - 1;
  z = (source_point->z / CUBE_SIDE) - 1;
  d = x + 3;
  e = y + 3;
  f = z + 3;

  if (x < 0)
    x = 0;
  if (y < 0)
    y = 0;
  if (z < 0)
    z = 0;
  if (d > world_x)
    d = world_x;
  if (e > world_y)
    e = world_y;
  if (f > world_z)
    f = world_z;

  /* collect the possible colliders from world objects */
  w = world_map;
  for (a = z; a < f; a++) {
    for (b = y; b < e; b++) {
      for (c = x; c < d; c++) {
	o = w[c + b*world_x + a*world_y*world_x];
	if (o != NULL) {
	  o->next_ext = NULL;
	  if (colliders_o == NULL) {
	    colliders_o = o;
	    p = o;
	  }
	  else {
	    p->next_ext = o;
	    p = o;
	  }
	}
      }
    }
  }

  /* collect the possible colliders from game objects */
  h = game_object_map;
  for (a = z; a < f; a++) {
    for (b = y; b < e; b++) {
      for (c = x; c < d; c++) {
	g = h[c + b*world_x + a*world_y*world_x];
	while (g != NULL) {
	  /* is the object of colliding type? */
	  if ((g->status & GO_COLLISION) != 0) {
	    g->next_ext = colliders_g;
	    colliders_g = g;
	  }
	  g = g->next;
	}
      }
    }
  }
}
