
/* (c) 2001 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>

#include "defines.h"
#include "connection.h"
#include "../common/parse.h"
#include "../common/socket.h"


extern volatile int connection_status;
extern char session_passwd[256], server_name[256], world_name[256];
extern int server_port, client_port, actual_port, actual_socket;

/* world information (from the config file) */
extern unsigned int world_seed, world_map;



int connect_to_host(const char *hostname, int p, int s) {

  struct sockaddr_in name;
  struct hostent *hostinfo;


  name.sin_family = AF_INET;
  name.sin_port = htons(p);
  hostinfo = gethostbyname(hostname);
  if (hostinfo == NULL) {
      fprintf (stderr, "CONNECT_TO_HOST: Unknown host %s.\n", hostname);
      return FAILED;
  }
  name.sin_addr = *(struct in_addr *)hostinfo->h_addr;

  if (connect(s, (struct sockaddr *)&name, sizeof(name)) < 0) {
    fprintf(stderr, "CONNECT_TO_HOST: Connect to %s:%d failed.\n", hostname, p);
    return FAILED;
  }

  fprintf(stderr, "CONNECT_TO_HOST: Connected to %s:%d.\n", hostname, p);
  return SUCCEEDED;
}


int connect_to_server(void) {

  unsigned int i;
  int s, size, n;
  char b[256];


  connection_status = OFF;

  while (1) {
    generate_port(&client_port);
    if (socket_create(&s, client_port, NO) == SUCCEEDED)
      break;
  }

  if (connect_to_host(server_name, server_port, s) == FAILED)
    return FAILED;

  sprintf(b+4, "%s%c", session_passwd, 0);
  size = 4 + strlen(b+4) + 1;
  b[size] = world_map;
  size++;
  write_int(b, size);
  write_int_a(b, &size, world_seed);

  send(s, b, size, 0);

  /* read the message size */
  size = recv(s, b, 256, 0);
  while (size < 4)
    size += recv(s, b+size, 256-size, 0);

  /* read the whole message */
  read_int(b, &n);
  while (size < n)
    size += recv(s, b+size, 256-size, 0);

  socket_delete(s);

  /* was the login successful? */
  if (b[4] == 0) {
    fprintf(stderr, "CONNECT_TO_SERVER: Could not login: Password - ");
    if (b[5] == 0)
      fprintf(stderr, "OK *** ");
    else
      fprintf(stderr, "WRONG *** ");

    fprintf(stderr, "World generation - ");
    if (b[6] == world_map)
      fprintf(stderr, "OK *** ");
    else {
      fprintf(stderr, "WRONG (");
      if (b[6] == WORLD_MAP_GENERATED)
	fprintf(stderr, "Should be generated using a seed value) *** ");
      else
	fprintf(stderr, "Should be generated using an image) *** ");
    }

    read_int(b+7, (int *)&i);
    fprintf(stderr, "Seed - ");
    if (i == world_seed)
      fprintf(stderr, "OK.\n");
    else
      fprintf(stderr, "WRONG (Server: %d).\n", i);

    fprintf(stderr, "CONNECT_TO_SERVER: Please fix these errors and try again.\n");

    return FAILED;
  }

  /* connection ok, got a free port! */
  n = 5;
  read_int_a(b, &n, &actual_port);
  sscanf(b+n, "%s", world_name);

  while (1) {
    generate_port(&client_port);
    if (socket_create(&actual_socket, client_port, NO) == SUCCEEDED)
      break;
  }

  if (connect_to_host(server_name, actual_port, actual_socket) == FAILED)
    return FAILED;

  connection_status = ON;

  return SUCCEEDED;
}
