
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "defines.h"
#include "game_object.h"


struct game_object **game_object_map = NULL, **game_object_table = NULL;
struct game_object **game_object_dummy_table = NULL;

extern struct game_object player_object;
extern int world_x, world_y, world_z;


int game_object_init(void) {

  int i;


  game_object_map = malloc(sizeof(struct game_object *) * world_x * world_y * world_z);
  game_object_table = malloc(sizeof(struct game_object *) * MAX_ID);
  game_object_dummy_table = malloc(sizeof(struct game_object *) * DUMMY_OBJECTS);
  if (game_object_map == NULL || game_object_table == NULL || game_object_dummy_table == NULL) {
    fprintf(stderr, "GAME_OBJECT_INIT: Out of memory.\n");
    return FAILED;
  }

  for (i = 0; i < world_x*world_y*world_z; i++)
    game_object_map[i] = NULL;

  for (i = 0; i < MAX_ID; i++)
    game_object_table[i] = NULL;

  for (i = 0; i < DUMMY_OBJECTS; i++)
    game_object_dummy_table[i] = NULL;

  return SUCCEEDED;
}


void game_object_free(struct game_object *g) {

  if (g == NULL)
    return;

  if (g->rotated_vertices != NULL)
    free(g->rotated_vertices);
  free(g);
}


void game_object_remove_all(void) {

  int i;


  for (i = 0; i < world_x*world_y*world_z; i++)
    game_object_map[i] = NULL;

  for (i = 0; i < MAX_ID; i++) {
    if (game_object_table[i] != NULL) {
      game_object_free(game_object_table[i]);
      game_object_table[i] = NULL;
    }
  }

  for (i = 0; i < DUMMY_OBJECTS; i++) {
    if (game_object_dummy_table[i] != NULL) {
      game_object_free(game_object_dummy_table[i]);
      game_object_dummy_table[i] = NULL;
    }
  }
}


int game_object_dummy_get(struct game_object **g) {

  int i;


  for (i = 0; i < DUMMY_OBJECTS; i++) {
    if (game_object_dummy_table[i] != NULL) {
      *g = game_object_dummy_table[i];
      game_object_dummy_table[i] = NULL;
      return SUCCEEDED;
    }
  }

  return FAILED;
}


int game_object_dummy_add(int id) {

  struct game_object *g;
  int i;


  for (i = 0; i < DUMMY_OBJECTS; i++) {
    if (game_object_dummy_table[i] == NULL) {

      g = malloc(sizeof(struct game_object));
      if (g == NULL) {
	fprintf(stderr, "GAME_OBJECT_DUMMY_ADD: Out of memory.\n");
	return FAILED;
      }

      g->id = id;
      g->rotated_vertices = NULL;
      g->next = NULL;
      g->prev = NULL;
      g->next_ext = NULL;

      game_object_dummy_table[i] = g;
      return SUCCEEDED;
    }
  }

  fprintf(stderr, "GAME_OBJECT_DUMMY_ADD: No room for ID %d.\n", id);

  return FAILED;
}


int game_object_create(char *name, int id, int object, double x, double y, double z, float ax, float ay, float az, struct game_object **o, int c) {

  struct game_object *g;


  if (name == NULL || o == NULL)
    return FAILED;

  g = malloc(sizeof(struct game_object));
  if (g == NULL) {
    fprintf(stderr, "GAME_OBJECT_CREATE: Out of memory.\n");
    return FAILED;
  }

  strcpy(g->name, name);
  g->id = id;
  g->object = object;
  g->x = x;
  g->y = y;
  g->z = z;
  g->ax = ax;
  g->ay = ay;
  g->az = az;
  g->rotated_vertices = NULL;
  g->status = 0;

  if (id != -1)
    game_object_add(g, c);

  *o = g;

  return SUCCEEDED;
}


int game_object_add(struct game_object *g, int c) {

  int x, y, z, i;


  if (g->id < 0 || g->id >= MAX_ID) {
    fprintf(stderr, "GAME_OBJECT_ADD: ID %d is out of range.\n", g->id);
    return FAILED;
  }

  if (game_object_table[g->id] != NULL) {
    fprintf(stderr, "GAME_OBJECT_ADD: Duplicate ID %d.\n", g->id);
    return FAILED;
  }

  /* toggle collision bit */
  if (c != 0)
    g->status = GO_COLLISION | GO_ROT_DIRTY;
  else
    g->status = GO_ROT_DIRTY;

  x = g->x / CUBE_SIDE;
  y = g->y / CUBE_SIDE;
  z = g->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  g->rotated_vertices = NULL;
  g->cx = x;
  g->cy = y;
  g->cz = z;
  i = z*world_y*world_x + y*world_x + x;

  g->next = game_object_map[i];
  g->next_ext = NULL;
  g->prev = NULL;

  if (game_object_map[i] != NULL)
    game_object_map[i]->prev = g;
  game_object_map[i] = g;
  game_object_table[g->id] = g;

  /* init sdi */
  g->sdi.type = DI_TYPE_GO;
  g->sdi.visibility_counter = 0;
  g->sdi.visibility_status = VIS_IS_HIDDEN;
  g->sdi.data = (void *)g;

  /* init mit */
  g->mit.ax = 0;
  g->mit.ay = 0;
  g->mit.az = 0;
  g->mit.vx = 0;
  g->mit.vy = 0;
  g->mit.vz = 0;
  g->mit.go = g;

  return SUCCEEDED;
}


int game_object_remove(int id) {

  struct game_object *g;
  int i;


  if (id < 0 || id > MAX_ID) {
    fprintf(stderr, "GAME_OBJECT_REMOVE: ID %d is out of range.\n", id);
    return FAILED;
  }

  g = game_object_table[id];

  if (g == NULL) {
    fprintf(stderr, "GAME_OBJECT_REMOVE: Missing object %d.\n", id);
    return FAILED;
  }

  if (g->prev == NULL) {
    i = g->cz*world_y*world_x + g->cy*world_x + g->cx;
    if (game_object_map[i] != g) {
      fprintf(stderr, "GAME_OBJECT_REMOVE: Object %d is in wrong place!\n", g->id);
      return FAILED;
    }
    game_object_map[i] = g->next;
    if (game_object_map[i] != NULL)
      game_object_map[i]->prev = NULL;
  }
  else {
    g->prev->next = g->next;
    if (g->next != NULL)
      g->next->prev = g->prev;
  }

  game_object_free(g);
  game_object_table[id] = NULL;

  return SUCCEEDED;
}


int game_object_get(int id, struct game_object **g) {

  if (id < 0 || id > MAX_ID) {
    fprintf(stderr, "GAME_OBJECT_GET: ID %d is out of range.\n", id);
    *g = NULL;
    return FAILED;
  }

  *g = game_object_table[id];
  if (*g == NULL) {
    /*
    fprintf(stderr, "GAME_OBJECT_GET: Unknown object ID %d.\n", id);
    */
    return FAILED;
  }

  return SUCCEEDED;
}


void game_object_relocate(struct game_object *g) {

  unsigned int x, y, z, i;


  x = g->x / CUBE_SIDE;
  y = g->y / CUBE_SIDE;
  z = g->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  /* is the object still in the same cube? */
  if (g->cx == x && g->cy == y && g->cz == z)
    return;

  /* cube changed -> relocate the object */
  if (g->next != NULL)
    g->next->prev = g->prev;
  if (g->prev != NULL)
    g->prev->next = g->next;
  else {
    i = g->cz*world_y*world_x + g->cy*world_x + g->cx;
    game_object_map[i] = g->next;
  }

  g->cx = x;
  g->cy = y;
  g->cz = z;    

  i = z*world_y*world_x + y*world_x + x;
  g->prev = NULL;
  g->next = game_object_map[i];
  if (game_object_map[i] != NULL)
    game_object_map[i]->prev = g;
  game_object_map[i] = g;
}
