
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sched.h>
#include <time.h>

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>

#include "defines.h"
#include "main.h"
#include "oload.h"
#include "txt.h"
#include "connection.h"
#include "game_object.h"
#include "ogl.h"
#include "gui.h"
#include "collision.h"
#include "time.h"
#include "font.h"
#include "console.h"
#include "world.h"
#include "network.h"
#include "particles.h"
#include "clouds.h"
#include "move_list.h"
#include "../common/parse.h"
#include "../common/socket.h"
#include "../common/glextd.h"
#include "../common/configuration.h"

#ifdef USE_DOF
#include "../common/dof.h"
#endif


extern struct txt_object *txt_first, *txt_last;
extern int screen_dx, screen_dy, fullscreen_mode;
extern int fps;

#ifdef INGAME_DEBUG
extern int draw_tri, cube_count, cloud_count, object_count;
#endif

struct game_object player_object, *skydome_object;
struct vector player_ellipsoid;

volatile int connection_status = OFF;
char session_passwd[256], server_name[256];
int server_port = 6789, client_port, actual_port, actual_socket;
int player_motion = 0, player_status = 0;
int today_hours = 20, today_minutes = 0, today_seconds = 0;
int priority = 0;

int status = 0;
char message[256], world_name[256];

int occlusion_available = NO, occlusion_status = OFF;

struct configuration configuration_client[] = {
  { "server:", CONFIGURATION_TYPE_STRING, server_name },
  { "uname:", CONFIGURATION_TYPE_STRING, player_object.name },
  { "passwd:", CONFIGURATION_TYPE_STRING, session_passwd },
  { "screendx:", CONFIGURATION_TYPE_INTEGER, &screen_dx },
  { "screendy:", CONFIGURATION_TYPE_INTEGER, &screen_dy },
  { "fullscreen:", CONFIGURATION_TYPE_INTEGER, &fullscreen_mode },
  { "priority:", CONFIGURATION_TYPE_INTEGER, &priority },
  { NULL, 0, NULL }
};

/* world information (from the config file) */
unsigned int world_seed = 0, world_map = 2;



int main(int argc, char *argv[]) {

  srand(time(NULL));

  /* set the defaults */
  strcpy(session_passwd, "candy!");
  strcpy(server_name, "localhost");
  strcpy(player_object.name, "cherry");

  /* read client.cfg */
  configuration_read("client.cfg", configuration_client);

  collision_init();
  txt_init("../data/txt/");
  network_init();

  /* write default values (if connection to server fails) */
  player_object.x = 180;
  player_object.y = 80;
  player_object.z = 180;
  player_object.ax = 0;
  player_object.ay = 0;
  player_object.az = 0;
  strcpy(world_name, "2");

  player_ellipsoid.x = 0.4;
  player_ellipsoid.y = 0.8;
  player_ellipsoid.z = 0.4;

  if (signal(SIGINT, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGKILL, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGTERM, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGSEGV, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGFPE, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGILL, signal_handler_int) == SIG_IGN)
    return 1;

  /* set the priority */
  setpriority(PRIO_PROCESS, 0, priority);

  if (read_world_cfg() == FAILED)
    return 1;

  if (connect_to_server() == FAILED)
    ;
  fprintf(stderr, "MAIN: Entering world \"%s\"...\n", world_name);

  init_window_main();

  world_init(world_name);
  move_list_init();
  game_object_init();
  font_init();
  console_init();
  object_load_list();
  particles_init();
  clouds_init();
  ogl_init();

  game_object_create("skydome", -1, OBJ_SKYDOME, 0, 0, 6, 0, 0, 0, &skydome_object, NO);

  main_window_set_title();
  init_main_glarea();

  /* get all data in one turn */
  if (connection_status == ON) {
    send_name();
    send_visibility();
    send_flush();
    read_server_data();
  }

  /* check if we have GL_HP_occlusion_test */
  if (glextd_is_supported("GL_HP_occlusion_test") == YES) {
    occlusion_available = YES;
  }

  SDL_ShowCursor(SDL_DISABLE);
  SDL_SetEventFilter(sdl_filter_events);

  /* GAME MAIN LOOP */
  while (1) {
    /* stupid 2.4.20+ kernel would run me so long without me yielding myself... */
    sched_yield();

    keyboard_io();
    compute_new_position();
    time_adjust_light(today_hours, today_minutes, today_seconds);
    /*
    position_moon(today_hours*60*60 + today_minutes*60 + today_seconds, moon_object);
    */
    move_list_pass();
    draw_main_glarea();
    communicate_with_server();
    if ((status & STATUS_QUIT) != 0 && connection_status == OFF)
      break;
  }

  txt_free_all();
#ifdef USE_DOF
  dof_exit();
#endif

  SDL_ShowCursor(SDL_ENABLE);
  SDL_Quit();

  return 0;
}


int read_world_cfg(void) {

  char tmp[256], c[256];
  FILE *f;


  strcpy(c, "../data/wld/");
  strcat(c, world_name);
  strcat(c, "/world.txt");

  f = fopen(c, "rb");
  if (f == NULL) {
    fprintf(stderr, "READ_WORLD_CFG: Could not open file \"%s\" for reading.\n", c);
    return FAILED;
  }

  while (fscanf(f, "%255s", tmp) != EOF) {
    if (strcmp(tmp, "MAP:") == 0) {
      fscanf(f, "%255s", tmp);
      if (strcmp(tmp, "GENERATED") == 0)
	world_map = WORLD_MAP_GENERATED;
      else if (strcmp(tmp, "IMAGE") == 0)
	world_map = WORLD_MAP_IMAGE;
      else
	fprintf(stderr, "READ_WORLD_CFG: File \"%s\", unknown symbol \"%s\".\n", c, tmp);
    }
    else if (strcmp(tmp, "SEED:") == 0) {
      fscanf(f, "%u", &world_seed);
    }
    else {
      fprintf(stderr, "READ_WORLD_CFG: File \"%s\", unknown symbol \"%s\".\n", c, tmp);
    }
  }

  fclose(f);

  return SUCCEEDED;
}


int position_moon(int t, struct game_object *m) {

  double d;


  /*
  d = t * (2*M_PI/(24*60*60));
  d -= 7*60*60 * (2*M_PI/(24*60*60));
  if (d < 0)
    d += 2*M_PI;

  s->x = player_object.x + cos(d)*SUN_DISTANCE;
  s->y = player_object.y + sin(d)*SUN_DISTANCE;
  s->z = player_object.z;

  d -= 12*60*60 * (2*M_PI/(24*60*60));
  if (d < 0)
    d += 2*M_PI;
  */

  /* the moon doesn't move */
  d = 3;

  m->x = player_object.x + cos(d)*MOON_DISTANCE;
  m->y = player_object.y + sin(d)*MOON_DISTANCE;
  m->z = player_object.z + 10;

  return SUCCEEDED;
}


void signal_handler_int(int signum) {

  if (connection_status == ON) {
    send_quit();
  }
  status |= STATUS_QUIT;
}


int process_client_command(char *b) {

  char t[256], tmp[256], x[4];
  int i;


  sscanf(b, "%s", t);

  /* x - close client */
  if (strcmp(t, "x") == 0) {
    send_quit();
    status |= STATUS_QUIT;
    console_write("Closing client...");
    return SUCCEEDED;
  }
  /* p - [none] - print position */
  else if (strcmp(t, "p") == 0) {
    sprintf(tmp, "(%.2f, %.2f, %.2f)", player_object.x, player_object.y, player_object.z);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* fps - [none] - print fps */
  else if (strcmp(t, "fps") == 0) {
    if (occlusion_status == ON)
      x[0] = 'O';
    else
      x[0] = '-';
    sprintf(tmp, "%d fps (%c)", fps, x[0]);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* t - [none] - print time */
  else if (strcmp(t, "t") == 0) {
    sprintf(tmp, "%.2d:%.2d:%.2d", today_hours, today_minutes, today_seconds);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* s - say something */
  else if (strcmp(t, "s") == 0) {
    if (connection_status == OFF) {
      console_write("Not connected.");
      return FAILED;
    }
    strcpy(message, b);
    return SUCCEEDED;
  }
  /* cl - close connection */
  else if (strcmp(t, "cl") == 0) {
    if (connection_status == OFF) {
      console_write("Not connected.");
      return FAILED;
    }
    send_quit();
    console_write("Closing connection...");
    return SUCCEEDED;
  }
  /* co - handshake - connect to server */
  else if (strcmp(t, "co") == 0) {
    if (connection_status == ON) {
      sprintf(tmp, "Already connected to %s:%d.", server_name, actual_port);
      console_write(tmp);
      return FAILED;
    }
    sprintf(tmp, "Connecting to %s:%d", server_name, actual_port);
    console_write(tmp);
    i = connect_to_server();
    if (i == SUCCEEDED) {
      console_write("SUCCEEDED.");
      send_name();
      send_visibility();
      send_flush();
      read_server_data();
      main_window_set_title();
      return SUCCEEDED;
    }
    console_write("FAILED.");
    return FAILED;
  }
  /* csn - [none] - change server name */
  else if (strcmp(t, "csn") == 0) {
    sscanf(b, "%*s %s", t);
    if (connection_status == ON) {
      sprintf(tmp, "Already connected to %s:%d.", server_name, actual_port);
      console_write(tmp);
      return FAILED;
    }
    strcpy(server_name, t);
    sprintf(tmp, "Changed server name to \"%s\".", server_name);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* e - [none] - empty console */
  else if (strcmp(t, "e") == 0) {
    console_empty();
    return SUCCEEDED;
  }
  /* cmp - [none] - change main port */
  else if (strcmp(t, "cmp") == 0) {
    sscanf(b, "%*s %d", &server_port);
    sprintf(tmp, "New main port %d.", server_port);
    console_write(tmp);
    return SUCCEEDED;
  }
  /* chp - [none] - change passwd */
  else if (strcmp(t, "chp") == 0) {
    sscanf(b, "%*s %s", t);
    if (connection_status == ON) {
      sprintf(tmp, "Already connected to %s:%d.", server_name, actual_port);
      console_write(tmp);
      return FAILED;
    }
    strcpy(session_passwd, t);
    sprintf(tmp, "New session passwd \"%s\".", session_passwd);
    console_write(tmp);
    return SUCCEEDED;
  }
#ifdef INGAME_DEBUG
  /* v - [none] - view game engine status */
  else if (strcmp(t, "v") == 0) {
    sprintf(tmp, "tris: %d cubes: %d clouds: %d objs: %d", draw_tri, cube_count, cloud_count, object_count);
    console_write(tmp);
    return SUCCEEDED;
  }
#endif
  /* cn - change name */
  else if (strcmp(t, "cn") == 0) {
    if (connection_status == OFF) {
      sscanf(b, "%*s %s", player_object.name);
      console_write("Not connected.");
      sprintf(tmp, "Your current _local_ name is \"%s\".", player_object.name);
      console_write(tmp);
      return SUCCEEDED;
    }
    else {
      sscanf(b, "%*s %s", t);
      if (strlen(t) == 0)
	return SUCCEEDED;
      sprintf(tmp, "N%s%c", t, 0);
      send_data(tmp, strlen(t) + 2);
    }
    return SUCCEEDED;
  }
  /* unknown command */
  else {
    console_write("Unknown command. Available commands:");

#ifdef INGAME_DEBUG
    console_write("cl, co, cn, csn, chp, e, fps, cmp, p, s, t, v, x");
#else
    console_write("cl, co, cn, csn, chp, e, fps, cmp, p, s, t, x");
#endif

    return FAILED;
  }

  return SUCCEEDED;
}
