
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <SDL/SDL_video.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "txt.h"
#include "ogl.h"
#include "gui.h"
#include "world.h"
#include "oload.h"
#include "particles.h"
#include "clouds.h"
#include "../common/frustum.h"
#include "../common/quaternion.h"
#include "../common/euler.h"
#include "../common/mesh2.h"

#ifdef USE_DOF
#include "../common/dof.h"
#endif


/* the active particle systems list */
extern struct particle_system *particle_systems;

/* the world particle systems map */
extern struct particle_system **particle_system_map;
extern struct particle_system **particle_system_table;

/* particle sorting list */
extern struct particle **particle_list;
extern int particle_list_size, particle_list_i;

extern struct object_object **world_map;
extern struct game_object player_object, *skydome_object;
extern struct game_object **game_object_map;
extern struct vector player_ellipsoid;
extern double player_x, player_y, today_light, today_fog;
extern int world_x, world_y, world_z, console_status, cross_status, console_txt, lantern_status, *object_type;
extern int occlusion_status, screen_dx, screen_dy;
extern char *object_occluder;
double cross_x, cross_y, cross_z;
float frustum[6*4], fov = FOV_MAX;
int show_blocks = 10;
int frame_count = 0, frame_number = 0, previous_ticks = 0, fps = 0, cube_count = 0, object_count = 0;
int current_texture_id;

struct game_object *closest_object = NULL;
struct vector player_source, player_target;

#ifdef INGAME_DEBUG
int draw_tri;
#endif

GLfloat light_ambient_max[] = { 1*1.5, 1*1.5, 0.9*1.5, 1.0 };
GLfloat light_ambient[] = { 0, 0, 0, 1.0 };
GLfloat light_diffuse_lantern[] = { 5, 5, 4, 1.0 };
GLfloat light_diffuse_moon[] = { 0.3, 0.3, 0.3, 1.0 };
GLfloat light_zero[] = { 0.0, 0.0, 0.0, 1.0 };
GLfloat light_position[] = { 0.0, 0.0, 0.0, 1.0 };
GLfloat color_sky_max[] = { 0, 0.5, 0.8 };
GLfloat color_sky[] = { 0, 0, 0 };
GLfloat color_lantern[] = { 5, 5, 4 };


struct draw_item **draw_item_array = NULL;
int draw_item_array_size = 1024*10;


int _ogl_draw_item_array_compare(const void *a, const void *b) {

  if ((*((struct draw_item **)a))->distance < (*((struct draw_item **)b))->distance)
    return -1;
  return 1;
}


int ogl_init(void) {

  draw_item_array = malloc(sizeof(struct draw_item *) * draw_item_array_size);
  if (draw_item_array == NULL) {
    fprintf(stderr, "OGL_INIT: Out of memory error.\n");
    return FAILED;
  }

  return SUCCEEDED;
}


int init_main_glarea(void) {

  glViewport(0, 0, screen_dx, screen_dy);

#ifdef USE_DOF
  /* init the depth of field computer */
  dof_init();
#endif

  return SUCCEEDED;
}


int _draw_skydome(void) {

  struct object_object *o;


  glPushMatrix();

  glTranslatef(player_object.x, player_object.y + player_ellipsoid.y/2, player_object.z);
  object_get(skydome_object->object, 1, &o);
  current_texture_id = -1;
  _render_object(o);

  glPopMatrix();

  return SUCCEEDED;
}


int draw_main_glarea(void) {

  struct particle_system *ps;
  struct object_object *o;
  struct txt_object *tx;
  struct game_object *g;
  double d, qx, qy, qz, lx, ly, lz;
  float dx, dy, dz;
  int x, y, z, ex, ey, ez, ax, ay, az, i;
  GLint vp[4];


#ifdef INGAME_DEBUG
  /* so far 0 triangles have been drawn this frame... */
  draw_tri = 0;
#endif

  /* update world blocks */
  world_dynamic_refresh();

  /* advance "time" */
  frame_count++;
  frame_number++;

  /* player normal: FIX THIS AND THE NEXT ONE TO USE QUATERNIONS */
  qx = 0;
  qy = 1;
  qz = 0;
  euler_rotate(&qx, &qy, &qz, M_PI*(-player_y)/180, 0, 0);
  euler_rotate(&qx, &qy, &qz, 0, 2*M_PI-player_x/180*M_PI, 0);

  /* player looks at */
  lx = 1;
  ly = 0;
  lz = 0;
  euler_rotate(&lx, &ly, &lz, M_PI*(-player_y)/180, 0, 0);
  euler_rotate(&lx, &ly, &lz, 0, 2*M_PI-player_x/180*M_PI, 0);

  player_source.x = player_object.x;
  player_source.y = player_object.y + player_ellipsoid.y/2;
  player_source.z = player_object.z;
  player_target.x = lx;
  player_target.y = ly;
  player_target.z = lz;

  /* save the direction normal vector */
  player_object.ax = qx;
  player_object.ay = qy;
  player_object.az = qz;

  lx += player_object.x;
  ly += player_object.y + player_ellipsoid.y/2;
  lz += player_object.z;

  /* refresh all changed textures */
  txt_refresh();

  /* opengl mode definitions */
  glClearColor(color_sky[0], color_sky[1], color_sky[2], 1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_TEXTURE_2D);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glFrontFace(GL_CCW);
  glEnable(GL_CULL_FACE);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_ambient);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthFunc(GL_LESS);

  light_diffuse_moon[0] = (1 - today_light)/2;
  light_diffuse_moon[1] = (1 - today_light)/2;
  light_diffuse_moon[2] = (1 - today_light)/2;

  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHT1);

  /* compute cross position */
  if (cross_status == ON) {
    cross_x = 1;
    cross_y = 0;
    cross_z = 0;
    euler_rotate(&cross_x, &cross_y, &cross_z, M_PI*(-player_y)/180, 0, 0);
    euler_rotate(&cross_x, &cross_y, &cross_z, 0, 2*M_PI-player_x/180*M_PI, 0);
    cross_x += player_object.x;
    cross_y += player_object.y + player_ellipsoid.y/2;
    cross_z += player_object.z;
    find_the_closest_pickable_object(cross_x, cross_y, cross_z, &closest_object, &d);
    if (d > 4)
      closest_object = NULL;
  }

  /* plot moonlight (directional) */
  dx = 100 - player_object.x;
  dy = 100 - player_object.y;
  dz = 100 - player_object.z;
  d = sqrt(dx*dx + dy*dy + dz*dz);
  light_position[0] = dx/d;
  light_position[1] = dy/d;
  light_position[2] = dz/d;
  light_position[3] = 0;
  glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.6);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.00001);
  glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.00001);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse_moon);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_zero);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glEnable(GL_LIGHT0);

  /* plot player's lantern light (positional) */
  if (lantern_status == ON) {
    refresh_lantern_light();
    light_position[0] = player_object.x;
    light_position[1] = player_object.y;
    light_position[2] = player_object.z;
    light_position[3] = 1;
    glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.3);
    glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.01);
    glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.1);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse_lantern);
    glLightfv(GL_LIGHT1, GL_SPECULAR, light_zero);
    glLightfv(GL_LIGHT1, GL_POSITION, light_position);
    glEnable(GL_LIGHT1);
  }

  /* distant places don't appear to be as clear as near by places */
  glFogf(GL_FOG_DENSITY, today_fog);
  glFogf(GL_FOG_MODE, GL_EXP);
  glFogf(GL_FOG_END, show_blocks*CUBE_SIDE);

  glGetIntegerv(GL_VIEWPORT, vp);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(fov, (double)vp[2]/(double)vp[3], 0.1, 1000);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(player_object.x, player_object.y + player_ellipsoid.y/2, player_object.z, lx, ly, lz, qx, qy, qz);

  glDisable(GL_LIGHTING);
  glDisable(GL_FOG);
  glBlendFunc(GL_SRC_COLOR, GL_ONE);

  /* draw the skydome */
  _draw_skydome();

  /* draw the clouds */
  glPushMatrix();
  clouds_manage();
  clouds_draw_all();
  glPopMatrix();

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(fov, (double)vp[2]/(double)vp[3], 0.1, show_blocks*CUBE_SIDE);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(player_object.x, player_object.y + player_ellipsoid.y/2, player_object.z, lx, ly, lz, qx, qy, qz);
  glPushMatrix();

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_FOG);

  /* obtain the current view frustum */
  frustum_get(frustum);

  /* draw the world */
  ax = (player_object.x / CUBE_SIDE) - show_blocks;
  ay = (player_object.y / CUBE_SIDE) - show_blocks;
  az = (player_object.z / CUBE_SIDE) - show_blocks;
  ex = ax + (show_blocks<<1) + 1;
  ey = ay + (show_blocks<<1) + 1;
  ez = az + (show_blocks<<1) + 1;

  /* don't handle cubes that are behind the viewer */
  z = get_player_x_segment();
  if (z == 0)
    ax += show_blocks - 2;
  else if (z == 1)
    az += show_blocks - 2;
  else if (z == 2)
    ex -= show_blocks - 1;
  else
    ez -= show_blocks - 1;

  if (ax < 0)
    ax = 0;
  if (ay < 0)
    ay = 0;
  if (az < 0)
    az = 0;
  if (ex > world_x)
    ex = world_x;
  if (ey > world_y)
    ey = world_y;
  if (ez > world_z)
    ez = world_z;

  /* build an array of all visible cubes and particle systems */
  particle_systems = NULL;
  i = 0;
  for (z = az; z < ez; z++) {
    for (y = ay; y < ey; y++) {
      for (x = ax; x < ex; x++) {
	if (world_map[z*world_x*world_y + y*world_x + x] != NULL) {
	  o = world_map[z*world_x*world_y + y*world_x + x];

	  if (o->sdi.visibility_counter < 0) {
	    o->sdi.visibility_counter++;
	  }
	  else {
	    if (o->sdi.visibility_counter == 0) {
	      if (frustum_test(&o->sdi) == NO) {
		o->sdi.visibility_counter = OC_FRAMES_INVISIBLE;
		o->sdi.visibility_status = VIS_IS_HIDDEN;
		continue;
	      }
	    }

	    /* compute player's squared distance to the midpoint */
	    dx = (x+0.5)*CUBE_SIDE - player_object.x;
	    dy = (y+0.5)*CUBE_SIDE - player_object.y;
	    dz = (z+0.5)*CUBE_SIDE - player_object.z;
	    o->sdi.distance = dx*dx + dy*dy + dz*dz;
	    draw_item_array[i++] = &o->sdi;
	    if (i == draw_item_array_size) {
	      printf("DRAW_MAIN_GLAREA: Out of object space!\n");
	      return FAILED;
	    }

	    /* add the voxel's particle systems to the list of systems */
	    if (o->sdi.distance < LOD_DISTANCE_PARTICLES) {
	      ps = particle_system_map[z*world_x*world_y + y*world_x + x];
	      while (ps != NULL) {
		ps->next_ext = particle_systems;
		particle_systems = ps;
		ps = ps->next;
	      }
	    }
	  }
	}
      }
    }
  }

  /* update the particles */
  particles_update();

  /* fill the z-buffer only */
  /*
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glShadeModel(GL_FLAT);
  glDisable(GL_FOG);
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);

  for (x = 0; x < i; x++)
    _render_object_plain_geometry(draw_item_array[x]->data);

  glEnable(GL_FOG);
  glEnable(GL_LIGHTING);
  glEnable(GL_TEXTURE_2D);
  glShadeModel(GL_SMOOTH);
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glDepthFunc(GL_LEQUAL);
  */

  /* build an array of all visible game objects */
  for (z = az; z < ez; z++) {
    for (y = ay; y < ey; y++) {
      for (x = ax; x < ex; x++) {
	g = game_object_map[z*world_x*world_y + y*world_x + x];
	while (g != NULL) {
	  if (g->status & GO_ROT_DIRTY)
	    game_object_transform(g);

	  if (g->sdi.visibility_counter < 0) {
	    g->sdi.visibility_counter++;
	  }
	  else {
	    if (g->sdi.visibility_counter == 0) {
	      if (frustum_test(&g->sdi) == NO) {
		g->sdi.visibility_counter = OC_FRAMES_INVISIBLE;
		g->sdi.visibility_status = VIS_IS_HIDDEN;
		g = g->next;
		continue;
	      }
	    }

	    dx = player_object.x - g->x;
	    dy = player_object.y - g->y;
	    dz = player_object.z - g->z;
	    g->sdi.distance = dx*dx + dy*dy + dz*dz;
	    draw_item_array[i++] = &g->sdi;
	    if (i == draw_item_array_size) {
	      printf("DRAW_MAIN_GLAREA: Out of object space!\n");
	      return FAILED;
	    }
	  }
	  g = g->next;
	}
      }
    }
  }

  /* sort the objects by distance to the player */
  qsort(draw_item_array, i, sizeof(struct draw_item *), _ogl_draw_item_array_compare);

  cube_count = 0;
  object_count = 0;

  /* draw the objects */
  current_texture_id = -1;
  for (x = 0; x < i; x++) {
    render_object(draw_item_array[x]);
  }

  /* draw the particles */
  glPopMatrix();
  glPushMatrix();
  particles_draw();

  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);

  /* cross (targeting) */
  if (cross_status == ON) {
    glPopMatrix();
    glPushMatrix();

    glDisable(GL_TEXTURE_2D);
    glColor4f(1, 1, 1, 1);
    glPointSize(10);
    glBegin(GL_POINTS);
    glVertex3f(cross_x, cross_y, cross_z);
    glEnd();
    glPointSize(1);

    /* target object */
    if (closest_object != NULL) {
      glPopMatrix();
      glPushMatrix();

      glBegin(GL_LINES);
      glVertex3f(closest_object->x, closest_object->y, closest_object->z);
      glVertex3f(cross_x, cross_y, cross_z);
      glEnd();
    }

    glEnable(GL_TEXTURE_2D);
  }

#ifdef USE_DOF
  /* compute the depth of field */
  dof_render();
#endif

  /* load HUD */
  if (console_status == ON) {

    glEnable(GL_TEXTURE_2D);

    glBlendFunc(GL_SRC_COLOR, GL_ONE);
    glEnable(GL_BLEND);
    glColor4f(1, 1, 1, 0.5);

    txt_get(console_txt, &tx);
    glBindTexture(GL_TEXTURE_2D, tx->ogl_name);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1, 1, -1, 1, 0, 0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity(); 

    glBegin(GL_QUADS);

    glTexCoord2f(0, 1);
    glVertex3f(-0.7, -0.9, 0);
    glTexCoord2f(1, 1);
    glVertex3f(0.7, -0.9, 0);
    glTexCoord2f(1, 0);
    glVertex3f(0.7, 0, 0);
    glTexCoord2f(0, 0);
    glVertex3f(-0.7, 0, 0);

    glEnd();

    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
  }

  /* still keep the player's viewpoint for lighting */
  glPopMatrix();

  /* swap buffers */
  SDL_GL_SwapBuffers();

  /* compute fps */
  x = SDL_GetTicks();
  if (x - previous_ticks > 1000) {
    fps = frame_count * 1000.0 / (x - previous_ticks);
    frame_count = 0;
    previous_ticks = x;
  }

  return SUCCEEDED;
}


int occlusion_test(float *aabb) {

  GLboolean result = GL_TRUE;


#ifdef GL_OCCLUSION_TEST_HP
  glDepthMask(GL_FALSE);
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glShadeModel(GL_FLAT);
  glDisable(GL_FOG);
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_CULL_FACE);
  glEnable(GL_OCCLUSION_TEST_HP);

  /* render the aabb */
  glPopMatrix();
  glPushMatrix();

  glBegin(GL_QUADS);

  /* front */
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MINY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MINY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MAXY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MAXY], aabb[AABB_MINZ]);

  /* back */
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MINY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MINY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MAXY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MAXY], aabb[AABB_MAXZ]);

  /* top */
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MAXY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MAXY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MAXY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MAXY], aabb[AABB_MAXZ]);

  /* bottom */
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MINY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MINY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MINY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MINY], aabb[AABB_MAXZ]);

  /* left */
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MINY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MAXY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MAXY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MINX], aabb[AABB_MINY], aabb[AABB_MAXZ]);

  /* right */
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MINY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MAXY], aabb[AABB_MINZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MAXY], aabb[AABB_MAXZ]);
  glVertex3f(aabb[AABB_MAXX], aabb[AABB_MINY], aabb[AABB_MAXZ]);

  glEnd();

  glDisable(GL_OCCLUSION_TEST_HP);
  glEnable(GL_FOG);
  glEnable(GL_LIGHTING);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_CULL_FACE);
  glShadeModel(GL_SMOOTH);
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glDepthMask(GL_TRUE);
  glGetBooleanv(GL_OCCLUSION_TEST_RESULT_HP, &result);
#endif

  return result;
}


int frustum_test(struct draw_item *di) {

  struct game_object *g;


  /* it's a plain object! */
  if (di->type == DI_TYPE_OO) {
    /* check if the object is inside the view frustum */
    if (frustum_sphere_inside(frustum, di->bounding_sphere[BS_X], di->bounding_sphere[BS_Y], di->bounding_sphere[BS_Z], di->bounding_sphere[BS_R]) == YES)
      return YES;
    return NO;
  }

  /* it's a game object! */
  g = (struct game_object *)di->data;

   /* check if the object is inside the view frustum */
  if (frustum_sphere_inside(frustum, g->x+di->bounding_sphere[BS_X], g->y+di->bounding_sphere[BS_Y], g->z+di->bounding_sphere[BS_Z], di->bounding_sphere[BS_R]) == YES)
    return YES;

  return NO;
}


int render_object(struct draw_item *di) {

  struct object_object *o;
  struct game_object *g;
  int i;


  if (di->visibility_counter < 0) {
    di->visibility_counter++;
    return SUCCEEDED;
  }

  /* it's a plain object! */
  if (di->type == DI_TYPE_OO) {
    if (di->visibility_counter > 0) {
      di->visibility_counter--;
    }
    else {
      /* do occlusion culling with occlusion test and aabb */
      if (occlusion_status == ON)
	i = occlusion_test(di->aabb);
      else
	i = GL_TRUE;

      /* return if the object was fully occluded */
      if (i == GL_FALSE) {
	di->visibility_counter = OC_FRAMES_INVISIBLE;
	di->visibility_status = VIS_IS_HIDDEN;
	return SUCCEEDED;
      }

      /* all visible objects are visible for the next n frames */
      di->visibility_counter = OC_FRAMES_VISIBLE;
    }

    cube_count++;
    glPopMatrix();
    glPushMatrix();
    di->visibility_status = VIS_IS_VISIBLE;
    _render_object((struct object_object *)di->data);

    return SUCCEEDED;
  }

  /* it's a game object! */
  g = (struct game_object *)di->data;

  if (di->visibility_counter > 0) {
    di->visibility_counter--;
  }
  else {
    /* do occlusion culling with occlusion test and aabb */
    if (occlusion_status == ON && object_occluder[g->object] != 0)
      i = occlusion_test(di->aabb);
    else
      i = GL_TRUE;

    /* return if the object was fully occluded */
    if (i == GL_FALSE) {
      di->visibility_counter = OC_FRAMES_INVISIBLE;
      di->visibility_status = VIS_IS_HIDDEN;
      return SUCCEEDED;
    }

    /* all visible objects are visible for the next n frames */
    di->visibility_counter = OC_FRAMES_VISIBLE;
  }

  glPopMatrix();
  glPushMatrix();
  glTranslatef(g->x, g->y, g->z);
  glMultMatrixf(g->rm);
  di->visibility_status = VIS_IS_VISIBLE;

  /* infer the object's lod */
  if (di->distance < LOD_1_DISTANCE)
    object_get(g->object, 1, &o);
  else
    object_get(g->object, 2, &o);

  if (o == NULL)
    return FAILED;

  object_count++;
  _render_object(o);

  return SUCCEEDED;
}


void _render_object(struct object_object *o) {

  struct mesh2 *s;
  struct txt_object *tx;
  int i, n, t, old_alpha, c;


  if (o == NULL)
    return;

  old_alpha = 0;
  s = o->mesh;

#ifdef INGAME_DEBUG
    draw_tri += s->triangles;
#endif

  glColor3f(1, 1, 1);

  glVertexPointer(3, GL_FLOAT, 0, s->vertex_list);
  glNormalPointer(GL_FLOAT, 0, s->vertex_normal_list);
  glTexCoordPointer(2, GL_FLOAT, 0, s->uv_list);

  c = 0;
  for (i = 0; i < s->objects; i++) {
    t = s->object_data[i*2 + 0];
    n = s->object_data[i*2 + 1];

    /* txt not loaded yet? */
    if (current_texture_id != t) {
      if (txt_get(t, &tx) == SUCCEEDED) {
	glBindTexture(GL_TEXTURE_2D, tx->ogl_name);
	/* do we have translucent triangles? */
	if (tx->status & TXT_STATUS_ALPHA) {
	  glEnable(GL_BLEND);
	  glDisable(GL_CULL_FACE);
	  glColor4f(1, 1, 1, tx->alpha);
	  old_alpha = 1;
	}
	else if (old_alpha == 1) {
	  glDisable(GL_BLEND);
	  glEnable(GL_CULL_FACE);
	  old_alpha = 0;
	}
	current_texture_id = t;
      }
    }

    /* draw the triangles */
    glDrawArrays(GL_TRIANGLES, c*3, n*3);
    c += n;
  }

  return;
}


void _render_object_plain_geometry(struct object_object *o) {

  struct mesh2 *s;
  int i, c;


  if (o == NULL)
    return;

  s = o->mesh;

#ifdef INGAME_DEBUG
  draw_tri += s->triangles;
#endif

  glVertexPointer(3, GL_FLOAT, 0, s->vertex_list);

  c = 0;
  for (i = 0; i < s->objects; i++) {
    c += s->object_data[i*2 + 1];
  }

  /* draw the triangles */
  glDrawArrays(GL_TRIANGLES, 0, c*3);

  return;
}


int find_the_closest_pickable_object(double x, double y, double z, struct game_object **go, double *cd) {

  struct game_object *g;
  double t, tx, ty, tz;
  int dx, dy, dz, a, b, c, d, e, f;


  *go = NULL;
  *cd = 10000;

  dx = (x / CUBE_SIDE) - 1;
  dy = (y / CUBE_SIDE) - 1;
  dz = (z / CUBE_SIDE) - 1;

  if (dx < 0)
    dx = 0;
  if (dy < 0)
    dy = 0;
  if (dz < 0)
    dz = 0;

  d = dx + 3;
  e = dy + 3;
  f = dz + 3;

  if (d > world_x)
    d = world_x;
  if (e > world_y)
    e = world_y;
  if (f > world_z)
    f = world_z;

  for (a = dz; a < f; a++) {
    for (b = dy; b < e; b++) {
      for (c = dx; c < d; c++) {
	g = game_object_map[a*world_x*world_y + b*world_x + c];
	while (g != NULL) {
	  if (object_type[g->object] == OBJECT_PICKABLE && g->sdi.visibility_status == VIS_IS_VISIBLE) {
	    tx = x - g->x;
	    ty = y - g->y;
	    tz = z - g->z;
	    t = tx*tx + ty*ty + tz*tz;
	    if (t < *cd) {
	      *go = g;
	      *cd = t;
	    }
	  }
	  g = g->next;
	}
      }
    }
  }

  return SUCCEEDED;
}


/* only DI_TYPE_GO's with GO_ROT_DIRTY set come here! */
void game_object_transform(struct game_object *g) {

  /* rotate the object if it needs (no old rotation matrix available) rotating */
  g->status &= ~GO_ROT_DIRTY;
  /* build the rotation matrix */
  quaternion_rotation_matrix(g->ax, g->ay, g->az, g->rm);
  /* transform vertices if the object affects collisions */
  if (g->status & GO_COLLISION)
    game_object_transform_object(g);
}


int game_object_transform_object(struct game_object *g) {

  struct object_object *o;
  struct draw_item *di;
  struct mesh2 *m;
  GLfloat *r;
  float *f, x, y, z;
  int n, i;


  if (g == NULL)
    return FAILED;

  object_get(g->object, 1, &o);
  if (o == NULL)
    return FAILED;

  m = o->mesh;
  di = &(g->sdi);

  /* has this object been transformed ever before? */
  if (g->rotated_vertices == NULL) {
    /* allocate room for transformed vertices */
    g->rotated_vertices = (float *)malloc(sizeof(float) * m->vertices*3);
    if (g->rotated_vertices == NULL) {
      fprintf(stderr, "GAME_OBJECT_TRANSFORM_OBJECT: Out of memory error.\n");
      return FAILED;
    }
  }

  /* rotate and translate the vertices */
  r = g->rm;

  /* init the aabb */
  f = m->vertex_list;
  di->aabb[AABB_MINX] = r[0]*f[0] + r[1]*f[1] + r[2]*f[2] + g->x;
  di->aabb[AABB_MAXX] = di->aabb[AABB_MINX];
  di->aabb[AABB_MINY] = r[4]*f[0] + r[5]*f[1] + r[6]*f[2] + g->y;
  di->aabb[AABB_MAXY] = di->aabb[AABB_MINY];
  di->aabb[AABB_MINZ] = r[8]*f[0] + r[9]*f[1] + r[10]*f[2] + g->z;
  di->aabb[AABB_MAXZ] = di->aabb[AABB_MINZ];

  for (i = 0, n = 0; n < m->vertices; n++) {
    /* rotate the vertex */
    g->rotated_vertices[i+0] = r[0]*f[i+0] + r[1]*f[i+1] + r[2]*f[i+2] + g->x;
    g->rotated_vertices[i+1] = r[4]*f[i+0] + r[5]*f[i+1] + r[6]*f[i+2] + g->y;
    g->rotated_vertices[i+2] = r[8]*f[i+0] + r[9]*f[i+1] + r[10]*f[i+2] + g->z;

    /* update the aabb */
    if (g->rotated_vertices[i+0] < di->aabb[AABB_MINX])
      di->aabb[AABB_MINX] = g->rotated_vertices[i+0];
    if (g->rotated_vertices[i+0] > di->aabb[AABB_MAXX])
      di->aabb[AABB_MAXX] = g->rotated_vertices[i+0];
    if (g->rotated_vertices[i+1] < di->aabb[AABB_MINY])
      di->aabb[AABB_MINY] = g->rotated_vertices[i+1];
    if (g->rotated_vertices[i+1] > di->aabb[AABB_MAXY])
      di->aabb[AABB_MAXY] = g->rotated_vertices[i+1];
    if (g->rotated_vertices[i+2] < di->aabb[AABB_MINZ])
      di->aabb[AABB_MINZ] = g->rotated_vertices[i+2];
    if (g->rotated_vertices[i+2] > di->aabb[AABB_MAXZ])
      di->aabb[AABB_MAXZ] = g->rotated_vertices[i+2];

    i += 3;
  }

  /* compute the bounding sphere */
  x = (di->aabb[AABB_MAXX] - di->aabb[AABB_MINX])/2;
  y = (di->aabb[AABB_MAXY] - di->aabb[AABB_MINY])/2;
  z = (di->aabb[AABB_MAXZ] - di->aabb[AABB_MINZ])/2;
  di->bounding_sphere[BS_X] = di->aabb[AABB_MINX] + x - g->x;
  di->bounding_sphere[BS_Y] = di->aabb[AABB_MINY] + y - g->y;
  di->bounding_sphere[BS_Z] = di->aabb[AABB_MINZ] + z - g->z;
  di->bounding_sphere[BS_R] = x*x + y*y + z*z;
  di->bounding_sphere[BS_RR] = sqrt(di->bounding_sphere[BS_R]);

  return SUCCEEDED;
}


void refresh_lantern_light(void) {

  float r;


  r = (rand()/(float)RAND_MAX)/2;
  light_diffuse_lantern[0] = color_lantern[0] + r;
  r = (rand()/(float)RAND_MAX)/2;
  light_diffuse_lantern[1] = color_lantern[1] + r;
  r = (rand()/(float)RAND_MAX)/2;
  light_diffuse_lantern[2] = color_lantern[2] + r;
}
