
#ifndef H_STRUCTURES
#define H_STRUCTURES

struct vector {
  float x, y, z;
};

#define DI_TYPE_OO 0
#define DI_TYPE_GO 1

#define BS_X  0
#define BS_Y  1
#define BS_Z  2
#define BS_R  3
#define BS_RR 4

#define VIS_IS_HIDDEN  0
#define VIS_IS_VISIBLE 1

struct draw_item {
  float distance;
  float aabb[3*2];          /* axis aligned bounding box for occlusion query */
  float bounding_sphere[5];
  int visibility_counter;   /* <0 - not visible, 0 - no idea, >0 - visible */
  int visibility_status;
  int type;
  void *data;
};

struct move_item {
  float ax, ay, az;         /* acceleration */
  float vx, vy, vz;         /* velocity */
  struct vector slide_plane_vector;
  struct game_object *go;
};

struct object_object {
  int display_list;
  int style;                /* 0 = no id, else id (BOTCLIENT) */
  struct draw_item sdi;     /* this object's draw item information */
  struct mesh2 *mesh;
  struct object_object *next_ext;
};

struct inventory_item {
  int object;
  struct inventory_item *next, *prev;
};

#define AABB_MINX 0
#define AABB_MAXX 1
#define AABB_MINY 2
#define AABB_MAXY 3
#define AABB_MINZ 4
#define AABB_MAXZ 5

struct game_object {
  int id;
  int object;
  int status;
  char name[64];
  double x, y, z;             /* position matrix */
  GLfloat rm[16];             /* rotation matrix */
  float ax, ay, az;           /* rotation */
  unsigned int cx, cy, cz;    /* cube indices */
  float *rotated_vertices;
  int hits;                   /* visible pixels (BOTCLIENT) */
  struct draw_item sdi;       /* draw item information */
  struct move_item mit;       /* move item information */
  struct game_object *next, *prev, *next_ext;
};

#define PARTICLES_MAX 64

struct particle {
  int age;
  float d;
  float sc;
  float x, y, z;
  float vx, vy, vz;
  float ax, ay, az;
  float cr, cg, cb;
};

#define PARTICLE_SYSTEM_POINT_FLAME 0
#define PARTICLE_SYSTEM_POINT_BLOOD 1
#define PARTICLE_SYSTEM_FACE_MIST   2

struct particle_system {
  int type;
  int id;
  int amount;
  int cx, cy, cz;
  float x, y, z;
  float dx, dy, dz;
  float *tri;
  struct particle *particles;
  struct particle_system *prev, *next, *next_ext;
};

struct cloud {
  int layer;
  float x, y, z;
  float alpha;
  float scale;
  float d;
  int txt_id;
  struct cloud *next;
};

#endif
