
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "defines.h"
#include "lsload.h"
#include "mesh.h"
#include "../client/txt.h"

/*
#define LS_DEBUG 1
*/

extern struct object_mesh *meshes_first, *meshes_last;
extern int txt_id_conversion;

int lsload_errors = YES;


/* e == YES -> show errors, e == NO -> don't show errors */
int object_load_show_errors(int e) {

  if (!(e == YES || e == NO))
    return FAILED;

  lsload_errors = e;

  return SUCCEEDED;
}


int object_load_landscape(char *name) {

  struct object_mesh *mesh;
  struct txt_object *txt;
  double *d;
  float *l;
  FILE *f;
  char t[256], mn[256];
  int m, i, *a, *b, vertices, normals, triangles, g, h, txt_list[256];


  f = fopen(name, "rb");
  if (f == NULL) {
    if (lsload_errors == YES)
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: Could not open file \"%s\".\n", name);
    return FAILED;
  }

#ifdef LS_DEBUG
  printf("FILE \"%s\"\n", name);
#endif

  meshes_first = NULL;
  meshes_last = NULL;

  /* find the amount of meshes */
  while (1) {
    fscanf(f, "%s", t);
    if (strcmp("Meshes:", t) == 0) {
      fscanf(f, "%d", &m);
      break;
    }
  }

  if (m == 0) {
    fclose(f);
    return SUCCEEDED;
  }

  /* get meshes */
  while (m > 0) {
    mesh = mesh_create();
    if (mesh == NULL)
      return FAILED;

    /* discard "// Mesh data" */
    fscanf(f, "%*s %*s %*s");

    /* mesh name */
    fscanf(f, "%s", t);
    if (t[0] == '"') {
      strcpy(mesh->name, &t[1]);
      mesh->name[strlen(mesh->name) - 1] = 0;
    }
    else
      strcpy(mesh->name, t);

#ifdef LS_DEBUG
    printf("mesh name: %s\n", mesh->name);
#endif

    /* vertices */
    fscanf(f, "%s", t);
    if (strcmp("Vertices:", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Vertices:\" found!\n", name);
      fclose(f);
      return FAILED;
    }
    fscanf(f, "%d", &vertices);

#ifdef LS_DEBUG
    printf("vertices: %d\n", vertices);
#endif

    /* normals */
    fscanf(f, "%s", t);
    if (strcmp("Normals:", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Normals:\" found!\n", name);
      fclose(f);
      return FAILED;
    }
    fscanf(f, "%s", t);
    if (strcmp("NO", t) == 0)
      normals = NO;
    else if (strcmp("YES", t) == 0)
      normals = YES;
    else {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no YES/NO found! Got \"%s\" instead.\n", name, t);
      fclose(f);
      return FAILED;
    }

#ifdef LS_DEBUG
    printf("normals: %s\n", t);
#endif

    /* triangles */
    fscanf(f, "%s", t);
    if (strcmp("Triangles:", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Triangles:\" found!\n", name);
      fclose(f);
      return FAILED;
    }
    fscanf(f, "%d", &triangles);

#ifdef LS_DEBUG
    printf("triangles: %d\n", triangles);
#endif

    /* textures and triangle list */
    mesh->textures = malloc(sizeof(int) * (triangles + triangles*3));
    if (mesh->textures == NULL) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", out of memory error.\n", name);
      free(mesh);
      return FAILED;
    }
    mesh->triangle_list = mesh->textures + triangles;

    mesh->uv_list = malloc(sizeof(float) * triangles * 2*3);
    mesh->vertice_list = malloc(sizeof(double) * vertices * 3);

    if (mesh->uv_list == NULL || mesh->vertice_list == NULL) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", out of memory error.\n", name);
      if (mesh->uv_list != NULL)
	free(mesh->uv_list);
      free(mesh->textures);
      free(mesh);
      return FAILED;
    }

    if (normals == YES) {
      mesh->vertex_normal_list = malloc(sizeof(float) * (vertices*3 + triangles*3));
      if (mesh->vertex_normal_list == NULL) {
	fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", out of memory error.\n", name);
	free(mesh->uv_list);
	free(mesh->textures);
	free(mesh->vertice_list);
	free(mesh);
	return FAILED;
      }
      mesh->face_normal_list = mesh->vertex_normal_list + vertices*3;
    }
    else {
      mesh->vertex_normal_list = NULL;
      mesh->face_normal_list = NULL;
    }

    mesh->next = NULL;
    mesh->vertices = vertices;
    mesh->triangles = triangles;

    /* discard "// Vertice data" */
    fscanf(f, "%*s %s %*s", t);
    if (strcmp("Vertice", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Vertice\" found! Got \"%s\" instead.\n", name, t);
      fclose(f);
      return FAILED;
    }

    d = mesh->vertice_list;
    for (i = 0; i < vertices; i++) {
      fscanf(f, "%lf %lf %lf", d, d+1, d+2);
      d += 3;
    }

    /* read vertex and face normals */
    if (normals == YES) {
      /* discard "// Vertex normals" */
      fscanf(f, "%*s %s %*s", t);
      if (strcmp("Vertex", t) != 0) {
	fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Vertex\" found! Got \"%s\" instead.\n", name, t);
	fclose(f);
	return FAILED;
      }

      l = mesh->vertex_normal_list;
      for (i = 0; i < vertices; i++) {
	fscanf(f, "%f %f %f", l, l+1, l+2);
	l += 3;
      }

      /* discard "// Face normals" */
      fscanf(f, "%*s %s %*s", t);
      if (strcmp("Face", t) != 0) {
	fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Face\" found! Got \"%s\" instead.\n", name, t);
	fclose(f);
	return FAILED;
      }

      l = mesh->face_normal_list;
      for (i = 0; i < triangles; i++) {
	fscanf(f, "%f %f %f", l, l+1, l+2);
	l += 3;
      }
    }

    /* discard "// Triangle data" */
    fscanf(f, "%*s %s %*s", t);
    if (strcmp("Triangle", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Triangle\" found! Got \"%s\" instead.\n", name, t);
      fclose(f);
      return FAILED;
    }

    a = mesh->triangle_list;
    b = mesh->textures;
    l = mesh->uv_list;
    for (i = 0; i < triangles; i++) {
      fscanf(f, "%d %d %d %d %f %f %f %f %f %f", b, a, a+1, a+2, l, l+1, l+2, l+3, l+4, l+5);
      b++;
      a += 3;
      l += 6;
    }

    /* discard "// Texture data" */
    fscanf(f, "%*s %s %*s", t);
    if (strcmp("Texture", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Texture\" found! Got \"%s\" instead.\n", name, t);
      fclose(f);
      return FAILED;
    }

    /* textures */
    fscanf(f, "%s", t);
    if (strcmp("Textures:", t) != 0) {
      fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", no \"Textures:\" found! Got \"%s\" instead.\n", name, t);
      fclose(f);
      return FAILED;
    }
    fscanf(f, "%d", &g);

    /* load textures */
    h = 0;
    while (g > 0) {
      /* texture name */
      fscanf(f, "%s", t);
      if (t[0] == '"') {
	strcpy(mn, &t[1]);
	mn[strlen(mn) - 1] = 0;
      }
      else
	strcpy(mn, t);

      /* load the texture */
      if (txt_load(mn, &i) == FAILED)
	txt_create(mn, &i, 64, 64);

      txt_list[h] = i;
      h++;
      if (txt_get(i, &txt) != SUCCEEDED) {
	fprintf(stderr, "OBJECT_LOAD_LANDSCAPE: File \"%s\", unknown texture \"%s\".\n", name, t);
	fclose(f);
	return FAILED;
      }

      g--;
    }

    /* convert local texture ids to local texture ids */
    if (txt_id_conversion == ON) {
      b = mesh->textures;
      for (i = 0; i < triangles; i++) {
	*b = txt_list[*b];
	b++;
      }
    }

    m--;

    /* add this mesh to the mesh list */
    if (meshes_first == NULL) {
      meshes_first = mesh;
      meshes_last = mesh;
    }
    else {
      meshes_last->next = mesh;
      meshes_last = mesh;
    }
  }

  fclose(f);

  return SUCCEEDED;
}
