
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "defines.h"
#include "../common/lsload.h"
#include "../common/parse.h"
#include "oload.h"
#include "txt.h"


struct object_object *objects_first = NULL, *objects_last = NULL;
struct object_mesh *meshes_first = NULL, *meshes_last = NULL, *mesh;
double sc_r, sc_g, sc_b, dc_r, dc_g, dc_b;
double ac_r, ac_g, ac_b, ec_r, ec_g, ec_b;
double *double_tmp, dob, *vertice_list;
double x_min, x_max, y_min, y_max, z_min, z_max;
double dx, dy, dz;
float *normal_list, *float_tmp;
char world_name[256], out_dir[256], mesh_name[256];
int double_tmp_size = 1024*8, txt_id, vertices, normals, triangles, *triangle_list;
int *int_tmp, int_tmp_size = 1024*8, materials_out[256], materials_out_count, float_tmp_size = 1024*8;
int cx, cy, cz;



int object_load_all(char *n) {

  char t[256], c[256], q[256], type[256];
  double x, y, z;
  FILE *f;


  double_tmp = malloc(sizeof(double) * double_tmp_size);
  if (double_tmp == NULL) {
    fprintf(stderr, "OBJECT_LOAD_ALL: Out of memory.\n");
    return FAILED;
  }
  int_tmp = malloc(sizeof(int) * int_tmp_size);
  if (int_tmp == NULL) {
    fprintf(stderr, "OBJECT_LOAD_ALL: Out of memory.\n");
    free(double_tmp);
    return FAILED;
  }
  float_tmp = malloc(sizeof(float) * float_tmp_size);
  if (float_tmp == NULL) {
    fprintf(stderr, "OBJECT_LOAD_ALL: Out of memory.\n");
    free(double_tmp);
    free(int_tmp);
    return FAILED;
  }

  strcpy(world_name, n);
  strcpy(t, "data/in/");
  strcat(t, n);
  strcat(t, "/");
  strcpy(c, t);
  strcat(t, "list.txt");
  strcpy(out_dir, "data/out/");
  strcat(out_dir, n);

  f = fopen(t, "rb");
  if (f == NULL) {
    fprintf(stderr, "OBJECT_LOAD_ALL: Could not open file \"%s\".\n", t);
    return FAILED;
  }

  while (1) {
    if (fscanf(f, "%s %s %lf %lf %lf", type, t, &x, &y, &z) == EOF)
      break;
    strcpy(q, c);
    strcat(q, t);
    if (object_load(q, x, y, z, type) == FAILED) {
      fclose(f);
      free(int_tmp);
      free(float_tmp);
      free(double_tmp);
      return FAILED;
    }
  }

  fclose(f);
  free(int_tmp);
  free(float_tmp);
  free(double_tmp);

  return SUCCEEDED;
}


int object_load(char *name, double x, double y, double z, char *type) {

  struct object_object *v;
  struct object_mesh *m;
  double *d;
  int s;


  /* landscape data? */
  if (strcmp("landscape", type) == 0) {
    if (object_load_landscape(name) == FAILED)
      return FAILED;
  }
  else {
    fprintf(stderr, "OBJECT_LOAD: Unknown object type \"%s\".\n", type);
    return FAILED;
  }

  v = malloc(sizeof(struct object_object));
  if (v == NULL) {
    fprintf(stderr, "OBJECT_LOAD: Out of memory.\n");
    return FAILED;
  }

  v->next = NULL;
  v->meshes = meshes_first;

  if (objects_first == NULL) {
    objects_first = v;
    objects_last = v;
  }
  else {
    objects_last->next = v;
    objects_last = v;
  }

  /* position meshes */
  m = v->meshes;
  while (m != NULL) {
    d = m->vertice_list;
    for (s = m->vertices; s > 0; s--) {
      *d += x;
      d++;
      *d += y;
      d++;
      *d += z;
      d++;
    }
    m = m->next;
  }

  return SUCCEEDED;
}


int object_get_next_token(char *b, char *r, int *s, int *k, int m) {

  while (*s < m && (*(b+*s) == ' ' || *(b+*s) == 0x0a || *(b+*s) == 0x0d))
    (*s)++;

  *k = 0;

  if (*(b+*s) == '"') {
    (*s)++;
    while (*s < m && *(b+*s) != '"') {
      *(r+*k) = *(b+*s);
      (*s)++;
      (*k)++;
    }
    (*s)++;
  }
  else {
    while (*s < m && *(b+*s) != ' ' && *(b+*s) != 0x0a && *(b+*s) != 0x0d) {
      *(r+*k) = *(b+*s);
      (*s)++;
      (*k)++;
    }
  }

  *(r+*k) = 0;

  return SUCCEEDED;
}


int position_triangles(void) {

  struct object_object *o;
  struct object_mesh *m;
  double x, y, z, *a;
  int i, *b, xx, yy, zz;


  printf(">> positioning triangles.\n");

  o = objects_first;
  if (o == NULL)
    return FAILED;
  while (o != NULL) {
    m = o->meshes;
    while (m != NULL) {
      m->triangle_cube = malloc(sizeof(int) * m->triangles);
      if (m->triangle_cube == NULL) {
	fprintf(stderr, "POSITION_TRIANGLES: Out of memory.\n");
	return FAILED;
      }
      a = m->triangle_center_xyz;
      b = m->triangle_cube;

      for (i = m->triangles; i > 0; i--) {
	x = *a - x_min;
	a++;
	y = *a - y_min;
	a++;
	z = *a - z_min;
	a++;
	xx = (int)(x / CUBE_SIDE);
	yy = (int)(y / CUBE_SIDE);
	zz = (int)(z / CUBE_SIDE);
	*b = zz*cy*cx + yy*cx + xx;
	b++;
      }

      m = m->next;
    }
    o = o->next;
  }

  return SUCCEEDED;
}


int compute_triangle_centers(void) {

  struct object_object *o;
  struct object_mesh *m;
  double x, y, z, *a, *c;
  int i, *b;


  printf(">> computing triangle centers.\n");

  o = objects_first;
  if (o == NULL)
    return FAILED;
  while (o != NULL) {
    m = o->meshes;
    while (m != NULL) {
      m->triangle_center_xyz = malloc(sizeof(double) * m->triangles * 3);
      if (m->triangle_center_xyz == NULL) {
	fprintf(stderr, "COMPUTE_TRIANGLE_CENTERS: Out of memory.\n");
	return FAILED;
      }
      a = m->triangle_center_xyz;
      b = m->triangle_list;
      c = m->vertice_list;

      for (i = m->triangles; i > 0; i--) {
	x = c[(*b * 3) + 0];
	y = c[(*b * 3) + 1];
	z = c[(*b * 3) + 2];
	b++;
	x += c[(*b * 3) + 0];
	y += c[(*b * 3) + 1];
	z += c[(*b * 3) + 2];
	b++;
	x += c[(*b * 3) + 0];
	y += c[(*b * 3) + 1];
	z += c[(*b * 3) + 2];
	b++;
	x /= 3;
	y /= 3;
	z /= 3;
	*(a++) = x;
	*(a++) = y;
	*(a++) = z;
      }

      m = m->next;
    }
    o = o->next;
  }

  return SUCCEEDED;
}


int find_dimensions(void) {

  struct object_object *o;
  struct object_mesh *m;
  double *d;
  int i, v = 0, me = 0, fa = 0;


  /* initialize values */
  o = objects_first;
  if (o == NULL)
    return FAILED;

  m = o->meshes;
  d = m->vertice_list;
  x_min = *d;
  x_max = *(d++);
  y_min = *d;
  y_max = *(d++);
  z_min = *d;
  z_max = *(d++);

  /* find cube dimensions */
  while (o != NULL) {
    m = o->meshes;
    while (m != NULL) {
      d = m->vertice_list;
      i = m->vertices;
      fa += m->triangles;
      v += i;
      me++;
      while (i > 0) {
	if (*d < x_min)
	  x_min = *d;
	if (*d > x_max)
	  x_max = *d;
	d++;
	if (*d < y_min)
	  y_min = *d;
	if (*d > y_max)
	  y_max = *d;
	d++;
	if (*d < z_min)
	  z_min = *d;
	if (*d > z_max)
	  z_max = *d;
	d++;
	i--;
      }
      m = m->next;
    }
    o = o->next;
  }

  dx = x_max - x_min;
  dy = y_max - y_min;
  dz = z_max - z_min;

  cx = (int)(dx / CUBE_SIDE) + 1;
  cy = (int)(dy / CUBE_SIDE) + 1;
  cz = (int)(dz / CUBE_SIDE) + 1;

  /* display world information */
  printf("WORLD \"%s\":\n", world_name);
  printf(" volume:    (%f, %f, %f): [%.2f, %.2f, %.2f] <--> [%.2f, %.2f, %.2f]\n", dx, dy, dz, x_min, y_min, z_min, x_max, y_max, z_max);
  printf(" vertices:  %d\n", v);
  printf(" meshes:    %d\n", me);
  printf(" triangles: %d\n", fa);

  return SUCCEEDED;
}


int create_out_dir(void) {

  mkdir(out_dir, 0);
  chmod(out_dir, S_IRUSR | S_IWUSR | S_IXUSR);

  return SUCCEEDED;
}


int write_world_data(void) {

  struct object_object *o;
  struct object_mesh *m;
  char t[256], c[256];
  int i, x, y, z, k, a, *p;
  FILE *f;


  printf(">> writing meshes.\n");

  strcpy(t, out_dir);
  strcat(t, "/info.txt");

  f = fopen(t, "wb");
  if (f == NULL) {
    fprintf(stderr, "WRITE_WORLD_DATA: Could not create file \"%s\".\n", t);
    return FAILED;
  }

  /* write world information */
  fprintf(f, "%d %d %d\n", cz, cy, cx);
  fclose(f);

  /* write cube files */
  i = 0;
  for (z = 0; z < cz; z++) {
    for (y = 0; y < cy; y++) {
      for (x = 0; x < cx; x++, i++) {
	sprintf(c, "%.3d-%.3d-%.3d.lan", z, y, x);
	printf("%.3d %.3d %.3d / %.3d %.3d %.3d\r", z, y, x, cz-1, cy-1, cx-1);
	strcpy(t, out_dir);
	strcat(t, "/");
	strcat(t, c);

	/* count meshes belonging to this cube */
	k = 0;
	o = objects_first;
	while (o != NULL) {
	  m = o->meshes;
	  while (m != NULL) {
	    p = m->triangle_cube;
	    for (a = m->triangles; a > 0; a--) {
	      if (*p == i) {
		k++;
		break;
	      }
	      p++;
	    }
	    m = m->next;
	  }
	  o = o->next;
	}

	/* no meshes? */
	if (k == 0)
	  continue;

	f = fopen(t, "wb");
	if (f == NULL) {
	  fprintf(stderr, "WRITE_WORLD_DATA: Could not create file \"%s\".\n", t);
	  return FAILED;
	}

	/* landscape 3d ascii header */
	fprintf(f, "// Landscape 3D ASCII\n");
	fprintf(f, "Meshes: %d\n", k);

	o = objects_first;
	while (o != NULL) {
	  m = o->meshes;
	  while (m != NULL) {
	    p = m->triangle_cube;
	    for (a = m->triangles; a > 0; a--) {
	      if (*p == i) {
		write_mesh(f, m, i);
		break;
	      }
	      p++;
	    }
	    m = m->next;
	  }
	  o = o->next;
	}

	fclose(f);
      }
    }
  }
  printf("\n");

  return SUCCEEDED;
}


int write_mesh(FILE *f, struct object_mesh *m, int i) {

  struct txt_object *t;
  static int *v = NULL, v_i = 0;
  double *d;
  float *l;
  int x, *y, *z, *p, a, b, c, *q, t_out, txt[256], no;


  if (m == NULL)
    return FAILED;

  t_out = 0;

  if (m->vertex_normal_list != NULL && m->face_normal_list != NULL)
    no = YES;
  else
    no = NO;

  if (v_i < m->vertices) {
    if (v != NULL)
      free(v);
    v = malloc(sizeof(int) * m->vertices * 2);
    v_i = m->vertices * 2;
  }

  /* count vertices, vertex normals and triangles */
  for (a = 0; a < m->vertices; a++)
    v[a] = 0xffffffff;

  p = m->triangle_cube;
  y = m->triangle_list;
  q = m->textures;
  for (a = 0, x = m->triangles; x > 0; x--) {
    if (*p == i) {
      /* handle texture */
      for (b = 0; b < t_out; b++)
	if (txt[b] == *q)
	  break;
      if (b == t_out) {
	txt[t_out] = *q;
	t_out++;
      }

      v[*(y+0)] = 0;
      v[*(y+1)] = 0;
      v[*(y+2)] = 0;
      a++;
    }
    p++;
    q++;
    y += 3;
  }

  /* reorder vertices */
  y = v;
  for (c = 0, x = m->vertices; x > 0; x--) {
    if (*y == 0) {
      *y = c;
      c++;
    }
    y++;
  }

  /* write mesh header */
  fprintf(f, "// Mesh data\n");
  fprintf(f, "\"%s\"\n", m->name);
  fprintf(f, "Vertices: %d\n", c);
  fprintf(f, "Normals: ");
  if (no == YES)
    fprintf(f, "YES\n");
  else
    fprintf(f, "NO\n");
  fprintf(f, "Triangles: %d\n", a);

  fprintf(f, "// Vertice data\n");
  d = m->vertice_list;
  for (x = 0; x < m->vertices; x++) {
    if (v[x] != 0xffffffff) {
      write_out_double(f, *(d+0), ' ');
      write_out_double(f, *(d+1), ' ');
      write_out_double(f, *(d+2), 0xA);
    }
    d += 3;
  }

  if (m->vertex_normal_list != NULL) {
    fprintf(f, "// Vertex normals\n");
    l = m->vertex_normal_list;
    for (x = 0; x < m->vertices; x++) {
      if (v[x] != 0xffffffff) {
	write_out_double(f, *(l+0), ' ');
	write_out_double(f, *(l+1), ' ');
	write_out_double(f, *(l+2), 0xA);
      }
      l += 3;
    }
  }

  if (m->face_normal_list != NULL) {
    fprintf(f, "// Face normals\n");
    l = m->face_normal_list;
    p = m->triangle_cube;
    for (x = 0; x < m->triangles; x++) {
      if (*p == i) {
	write_out_double(f, *(l+0), ' ');
	write_out_double(f, *(l+1), ' ');
	write_out_double(f, *(l+2), 0xA);
      }
      p++;
      l += 3;
    }
  }

  fprintf(f, "// Triangle data\n");
  z = m->triangle_list;
  l = m->uv_list;
  p = m->triangle_cube;
  for (x = 0; x < m->triangles; x++) {
    if (*p == i) {
      /* convert texture */
      for (b = 0; b < t_out; b++) {
	if (txt[b] == m->textures[x])
	  break;
      }
      fprintf(f, "%d %d %d %d ", b, v[*(z+0)], v[*(z+1)], v[*(z+2)]);
      write_out_double(f, *(l+0), ' ');
      write_out_double(f, *(l+1), ' ');
      write_out_double(f, *(l+2), ' ');
      write_out_double(f, *(l+3), ' ');
      write_out_double(f, *(l+4), ' ');
      write_out_double(f, *(l+5), 0xA);
    }
    p++;
    z += 3;
    l += 6;
  }

  fprintf(f, "// Texture data\n");
  fprintf(f, "Textures: %d\n", t_out);

  for (x = 0; x < t_out; x++) {
    if (txt_get(txt[x], &t) == FAILED)
      return FAILED;
    fprintf(f, "\"%s\"\n", t->name);
  }

  return SUCCEEDED;
}
