
/* (c) 2001-2002 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "defines.h"
#include "common.h"


extern double *map, *map_vertices_z, *map_vertices_xy, *map_normals, *map_triangle_normals;
extern char *map_shadow, *map_area;
extern int size, *map_triangles;


int map_insert(int x, int y, double d) {

  if (x < 0 || x >= size)
    return FAILED;
  if (y < 0 || y >= size)
    return FAILED;

  if (d < 0)
    d = 0;

  if (map_shadow[y * size + x] == 0) {
    map_shadow[y * size + x] = 1;
    map[y * size + x] = d;
    return SUCCEEDED;
  }

  return FAILED;
}


int map_insert_force(int x, int y, double d) {

  if (x < 0 || x >= size)
    return FAILED;
  if (y < 0 || y >= size)
    return FAILED;

  if (d < 0)
    d = 0;

  map_shadow[y * size + x] = 1;
  map[y * size + x] = d;
  return SUCCEEDED;
}


int map_read(int x, int y, double *d) {

  while (x < 0)
    x += size;
  while (y < 0)
    y += size;
  while (x >= size)
    x -= size;
  while (y >= size)
    y -= size;

  *d += map[y * size + x];
  return SUCCEEDED;
}


int map_area_read(int x, int y, char *c) {

  while (x < 0)
    x += size;
  while (y < 0)
    y += size;
  while (x >= size)
    x -= size;
  while (y >= size)
    y -= size;

  *c = map_area[y * size + x];
  return SUCCEEDED;
}


int is_point_inside_a_triangle(struct vector *a, struct vector *b, struct vector *c, struct vector *p) {

  struct vector an, bn, cn;
  double l, d;


  an.x = a->x - p->x;
  an.y = a->y - p->y;
  an.z = a->z - p->z;
  bn.x = b->x - p->x;
  bn.y = b->y - p->y;
  bn.z = b->z - p->z;
  cn.x = c->x - p->x;
  cn.y = c->y - p->y;
  cn.z = c->z - p->z;
  l = an.x*an.x + an.y*an.y + an.z*an.z;
  l = sqrt(l);
  an.x /= l;
  an.y /= l;
  an.z /= l;
  d = bn.x*bn.x + bn.y*bn.y + bn.z*bn.z;
  d = sqrt(d);
  bn.x /= d;
  bn.y /= d;
  bn.z /= d;
  l = cn.x*cn.x + cn.y*cn.y + cn.z*cn.z;
  l = sqrt(l);
  cn.x /= l;
  cn.y /= l;
  cn.z /= l;

  d = an.x*bn.x + an.y*bn.y + an.z*bn.z;
  if (d > 1)
    d = 1;
  else if (d < -1)
    d = -1;
  l = acos(d);

  d = bn.x*cn.x + bn.y*cn.y + bn.z*cn.z;
  if (d > 1)
    d = 1;
  else if (d < -1)
    d = -1;
  l += acos(d);

  d = an.x*cn.x + an.y*cn.y + an.z*cn.z;
  if (d > 1)
    d = 1;
  else if (d < -1)
    d = -1;
  l += acos(d);

  if (l >= (M_PI * 2) - EPSILON)
    return SUCCEEDED;
  return FAILED;
}
