
/* (c) 2001-2002 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtkvscrollbar.h>
#include <gdk/gdkkeysyms.h>

#include "GL/gl.h"
#include "../common/png.h"

#include "gui.h"
#include "defines.h"


GtkWidget *main_window, *main_draw_area, *main_vbox;
guchar *main_buffer;

extern double *map;
extern char *map_shadow, *map_min_max;
extern int size;


int gui_init(int argc, char *argv[]) {

  main_buffer = malloc(sizeof(guchar) * size * size * 3);
  if (main_buffer == NULL) {
    fprintf(stderr, "GUI_INIT: Out of memory error.\n");
    return FAILED;
  }

  map_generator_draw();

  /* gtk shit */
  gtk_set_locale();
  gtk_init(&argc, &argv);

  /* gdk rgd inits */
  gdk_rgb_init();
  gtk_widget_set_default_colormap(gdk_rgb_get_cmap());
  gtk_widget_set_default_visual(gdk_rgb_get_visual());

  /* memory window */
  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT(main_window), "destroy", GTK_SIGNAL_FUNC(destroy), NULL);

  /* vertical boxes */
  main_vbox = gtk_vbox_new(FALSE, 0);

  /* drawing area */
  main_draw_area = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(main_draw_area), size, size);
  gtk_signal_connect(GTK_OBJECT(main_draw_area), "expose_event", GTK_SIGNAL_FUNC(main_draw_area_expose), NULL);

  gtk_container_add(GTK_CONTAINER(main_window), main_vbox);
  gtk_box_pack_start(GTK_BOX(main_vbox), main_draw_area, FALSE, TRUE, 0);

  /* display */
  gtk_widget_show_all(main_window);
  gtk_main();

  return SUCCEEDED;
}


gboolean main_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data) {
  gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
                     0, 0, size, size, GDK_RGB_DITHER_NORMAL, main_buffer, size * 3);
  return FALSE;
}


void destroy(GtkWidget *widget, gpointer data) {

  gtk_main_quit();
}


static int _clamp_value(int v) {

  if (v < 0)
    return 0;
  if (v > 255)
    return 255;

  return v;
}


int map_generator_draw(void) {

  int x;


  /* grayscale -> for debugging */
  for (x = 0; x < size*size; x++) {
    main_buffer[x*3 + 0] = map[x];
    main_buffer[x*3 + 1] = map[x];
    main_buffer[x*3 + 2] = map[x];
  }

  png_save("tmp0.png", size, size, GL_RGB, main_buffer);

  for (x = 0; x < size*size; x++) {
    if (map_min_max[x] == 0) {
      /* sea? */
      if (map[x] < LIMIT_SEA) {
	main_buffer[x*3 + 0] = 10;
	main_buffer[x*3 + 1] = 30;
	main_buffer[x*3 + 2] = _clamp_value(40 + map[x]*2);
      }
      else if (map[x] < LIMIT_GRASS) {
	main_buffer[x*3 + 0] = 30;
	main_buffer[x*3 + 1] = _clamp_value(60 + map[x]*2);
	main_buffer[x*3 + 2] = 30;
      }
      else {
	main_buffer[x*3 + 0] = _clamp_value(30 + map[x]*2);
	main_buffer[x*3 + 1] = _clamp_value(30 + map[x]*2);
	main_buffer[x*3 + 2] = _clamp_value(30 + map[x]*2);
      }
    }
    else if (map_min_max[x] == MAP_MIN) {
      main_buffer[x*3 + 0] = 255;
      main_buffer[x*3 + 1] = 0;
      main_buffer[x*3 + 2] = 0;
    }
    else {
      main_buffer[x*3 + 0] = 255;
      main_buffer[x*3 + 1] = 255;
      main_buffer[x*3 + 2] = 255;
    }
  }

  png_save("tmp1.png", size, size, GL_RGB, main_buffer);

  for (x = 0; x < size*size; x++) {
    /* sea? */
    if (map[x] < LIMIT_SEA) {
      main_buffer[x*3 + 0] = 10;
      main_buffer[x*3 + 1] = 30;
      main_buffer[x*3 + 2] = _clamp_value(40 + map[x]*2);
    }
    else if (map[x] < LIMIT_GRASS) {
      main_buffer[x*3 + 0] = 30;
      main_buffer[x*3 + 1] = _clamp_value(60 + map[x]*2);
      main_buffer[x*3 + 2] = 30;
    }
    else {
      main_buffer[x*3 + 0] = _clamp_value(30 + map[x]*2);
      main_buffer[x*3 + 1] = _clamp_value(30 + map[x]*2);
      main_buffer[x*3 + 2] = _clamp_value(30 + map[x]*2);
    }
  }

  png_save("tmp2.png", size, size, GL_RGB, main_buffer);

  return SUCCEEDED;
}
