
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <GL/gl.h>

#include "defines.h"
#include "gui.h"
#include "main.h"
#include "build.h"
#include "smooth.h"
#include "common.h"
#include "../common/random.h"
#include "../common/png.h"
#include "../common/hash.h"


static int _save_world_information(char *name, unsigned int seed);


/* define this if you want to output normals */
#define NORMALS 1


struct plant **obj_plants = NULL;
struct psystem *psystems = NULL;

double *map = NULL, *map_vertices_z = NULL, *map_vertices_xy = NULL;
double random_xy = DEF_RANDOM_XY, limit = DEF_LIMIT, divisor = DEF_DIVISOR;
float *map_normals = NULL, *map_triangle_normals = NULL, *map_fertility = NULL;
float *map_slopes = NULL;
char *map_shadow = NULL, *map_area = NULL, *map_triangle_area = NULL, input_file[256], *map_min_max = NULL;
int size = MAP_SIZE, *map_triangles = NULL, gui_status = ON, write_status = ON;

/* image file */
int idx, idy, itype;
unsigned char *ic;

/* houses */
struct house *houses = NULL;

/* seed */
unsigned int seed = 0;



int main(int argc, char *argv[]) {

  unsigned int s;


  s = time(NULL);
  input_file[0] = 0;

  if (parse_arguments(argc, argv, &s) == FAILED) {
    printf("USAGE: %s [--nogui] [--nowrite] [--seed seed] [--file file]\n", argv[0]);
    return 0;
  }

  seed = s;

  /* load the input file if it was given */
  if (input_file[0] != 0) {
    if (map_load_png(input_file) == FAILED)
      return 1;
    size = idx;
  }

  /* save the world information */
  if (_save_world_information(input_file, s) == FAILED)
    return 1;

  /* allocate memory */
  map_vertices_z = malloc(sizeof(double) * (size*2 + 1) * (size*2 + 1));
  map_vertices_xy = malloc(sizeof(double) * (size*2 + 1) * (size*2 + 1) * 2);
  map = malloc(sizeof(double) * size*size);
  map_shadow = malloc(size*size);
  map_area = malloc(size*size);
  map_normals = malloc(sizeof(float) * (size*2 + 1) * (size*2 + 1) * 3);
  map_fertility = malloc(sizeof(float) * size*size);
  map_slopes = malloc(sizeof(float) * size*size*8);
  map_triangles = malloc(sizeof(int) * size*size*8*3);
  map_triangle_area = malloc(size*size*8);
  map_triangle_normals = malloc(sizeof(float) * size*size*8*3);
  map_min_max = malloc(size*size);
  obj_plants = malloc(sizeof(struct plant *) *size*size*8);
  if (map == NULL || map_shadow == NULL || map_area == NULL || map_vertices_z == NULL || map_fertility == NULL ||
      map_vertices_xy == NULL || map_normals == NULL || map_triangles == NULL || map_min_max == NULL ||
      map_triangle_normals == NULL || obj_plants == NULL || map_triangle_area == NULL) {
    fprintf(stderr, "MAIN: Out of memory error.\n");
    return FAILED;
  }

  memset(map, 0, sizeof(double) * size*size);
  memset(map_fertility, 0, sizeof(float) * size*size);
  memset(map_shadow, 0, size*size);
  memset(map_min_max, 0, size*size);
  memset(map_area, MAP_AREA_UNDEFINED, sizeof(char) * size*size);
  memset(obj_plants, 0, sizeof(struct plant *) * size*size*8);

  /* create the map */
  if (input_file[0] != 0) {
    if (map_convert_png() == FAILED)
      return 1;
  }
  else {
    if (map_generate(s) == FAILED)
      return 1;
  }

  map_generator_insert_houses();

  map_generator_find_local_minima();
  map_generator_spread_minima();
  map_generator_compute_fertility();

  map_generator_build_areas();
  remove_lonely_area_squares();
  map_generator_build_vertices();
  map_generator_build_triangles();
  map_generator_build_normals();
  map_generator_compute_slopes();
  map_generator_plant_extras();
  map_generator_plant_psystems();

  if (write_status == ON) {
    map_generator_save_mesh();
    map_generator_save_plants();
    map_generator_save_psystems();
    map_generator_save_houses();
  }

  /* display */
  if (gui_status == ON)
    gui_init(argc, argv);

  return 0;
}


static int _save_world_information(char *name, unsigned int seed) {

  unsigned int s;
  FILE *f;


  f = fopen("world.txt", "wb");
  if (f == NULL) {
    fprintf(stderr, "SAVE_WORLD_INFORMATION: Could not open file \"world.txt\" for writing.\n");
    return FAILED;
  }

  if (name[0] == 0) {
    fprintf(f, "MAP: GENERATED\n");
    fprintf(f, "SEED: %d\n", seed);
  }
  else {
    fprintf(f, "MAP: IMAGE\n");
    if (hash_file(name, &s) == FAILED)
      return FAILED;
    fprintf(f, "SEED: %d\n", s);
  }

  fclose(f);

  return SUCCEEDED;
}


int map_convert_png(void) {

  int x, y;


  for (y = 0; y < idy; y++) {
    for (x = 0; x < idx; x++) {
      if (itype == GL_RGBA)
	map_insert(x, y, ((double)ic[y*idx*4 + x*4 + 0])/10.0);
      else
	map_insert(x, y, ((double)ic[y*idx*3 + x*3 + 0])/10.0);
    }
  }

  return SUCCEEDED;
}


int map_load_png(char *name) {

  if (png_load(name, &idx, &idy, &itype, &ic) == FAILED)
    return FAILED;

  if (idx != idy) {
    fprintf(stderr, "MAP_LOAD_PNG: The image \"%s\" size is %dx%d, so it's not a square.\n", name, idx, idy);
    return FAILED;
  }

  fprintf(stderr, "MAP_LOAD_PNG: Image size %dx%d.\n", idx, idy);

  return SUCCEEDED;
}


int parse_arguments(int argc, char *argv[], unsigned int *s) {

  int i;


  i = 1;
  while (i < argc) {
    if (strcmp(argv[i], "--nogui") == 0)
      gui_status = OFF;
    else if (strcmp(argv[i], "--file") == 0) {
      i++;
      strcpy(input_file, argv[i]);
    }
    else if (strcmp(argv[i], "--seed") == 0) {
      i++;
      *s = atoi(argv[i]);
    }
    else if (strcmp(argv[i], "--nowrite") == 0) {
      write_status = OFF;
    }
    else
      return FAILED;

    i++;
  }

  return SUCCEEDED;
}


int map_generate(unsigned int s) {

  double d, t;
  int r, x = 1, y = 1, a, b, g, q, p;


  srand(s);
  printf("seed: %d\n", s);

  r = size - 1;

  /* init borders */
  for (a = 0; a < size; a++) {
    map_insert(a, 0, 0);
    map_insert(0, a, 0);
    map_insert(a, size - 1, 0);
    map_insert(size - 1, a, 0);
  }

  /* first step */
  d = map[0] + map[size - 1] + map[size * (size - 1)] + map[size * (size - 1) + size - 1];
  d /= 4;
  d += random_gaussian(limit);
  map_insert((size - 1) / 2, (size - 1) / 2, d);

  x = x<<1;
  y = y<<1;

  /* generate terrain - uses the diamond-square -algorithm
     (see e.g., http://www.gameprogrammer.com/fractal.html) */
  while (r > 1) {
    /* diamond steps */
    g = r>>1;
    for (q = 0, b = 0; b < y; b++, q += r) {
      for (p = 0, a = 0; a < x; a++, p += r) {
	/* right */
	d = 0;
	map_read(p + 0, q + 0, &d);
	map_read(p + r, q + 0, &d);
	map_read(p + g, q + g, &d);
	map_read(p + g, q - g, &d);
	d /= 4;
	d += random_gaussian(limit);
	map_insert(p + g, q, d);
	/* down */
	d = 0;
	map_read(p + 0, q + 0, &d);
	map_read(p + 0, q + r, &d);
	map_read(p + g, q + g, &d);
	map_read(p - g, q + g, &d);
	d /= 4;
	d += random_gaussian(limit);
	map_insert(p, q + g, d);
      }
    }

    r = r>>1;

    if (r <= 1)
      break;

    /* square steps */
    g = r>>1;
    for (q = 0, b = 0; b < y; b++, q += r) {
      for (p = 0, a = 0; a < x; a++, p += r) {
	d = 0;
	map_read(p + 0, q + 0, &d);
	map_read(p + r, q + 0, &d);
	map_read(p + r, q + r, &d);
	map_read(p + 0, q + r, &d);
	d /= 4;
	d += random_gaussian(limit);
	map_insert(p + g, q + g, d);
      }
    }

    x = x<<1;
    y = y<<1;

    /* diminish the effect of the random generator as iterations go on */
    limit /= divisor;
  }

  /* bring the negative bottom floor to zero level */
  for (b = map[0], t = map[0], x = 1; x < size*size; x++) {
    if (map[x] < t)
      t = map[x];
    if (map[x] > b)
      b = map[x];
  }

  for (x = 0; x < size*size; x++) {
    map[x] += -t;
  }

  /* smooth the result */
  smooth_run();

  return SUCCEEDED;
}


int map_generator_print(void) {

  int a, b;


  for (b = 0; b < size; b++) {
    for (a = 0; a < size; a++) {
      printf("%0.1f ", map[b * size + a]);
    }
    printf("\n");
  }
  printf("******************* shadow ***********************\n");
  for (b = 0; b < size; b++) {
    for (a = 0; a < size; a++) {
      printf("%1.1d ", map_shadow[b * size + a]);
    }
    printf("\n");
  }
  printf("--------------------------------------------------\n");

  return SUCCEEDED;
}


int map_generator_save_binary(void) {

  double max;
  FILE *f;
  char c;
  int x, y;


  f = fopen("out", "wb");
  if (f == NULL)
    return FAILED;

  for (max = 0, x = 0; x < size * size; x++) {
    if (map[x] > max)
      max = map[x];
  }

  for (y = 0; y < size; y++) {
    for (x = 0; x < size; x++) {
      c = (char)(map[y * size + x] * 255 / max);
      fprintf(f, "%c%c%c", c, c, c);
    }
  }

  fclose(f);

  return SUCCEEDED;
}


int map_generator_save_mesh(void) {

  double dx, dy, dz, *d, *w;
#ifdef NORMALS
  float *s;
#endif
  float *ms;
  FILE *f;
  char *r;
  int a, b, q, x, y, *m, t, xi, p;


  f = fopen("landscape.lan", "wb");
  if (f == NULL) {
    fprintf(stderr, "MAP_GENERATOR_SAVE_MESH: Could not open file \"landscape.lan\" for writing.\n");
    return FAILED;
  }

  p = (size<<1) + 1;
  q = (size*size)<<3;

  /* print header */
  fprintf(f, "// Landscape 3D ASCII\n");
  fprintf(f, "Meshes: %d\n", 1);
  fprintf(f, "// Mesh data\n");
  fprintf(f, "\"world\"\n");
  fprintf(f, "Vertices: %d\n", p*p);
#ifdef NORMALS
  fprintf(f, "Normals: YES\n");
#else
  fprintf(f, "Normals: NO\n");
#endif
  fprintf(f, "Triangles: %d\n", q);

  /* output vertices */
  fprintf(f, "// Vertice data\n");
  d = map_vertices_z;
  w = map_vertices_xy;
  for (y = 0; y < p; y++) {
    for (x = 0; x < p; x++) {
      dz = *(d++) * MAP_PIXEL_SIDE_SCALE;
      dx = *(w++) * MAP_PIXEL_SIDE_SCALE;
      dy = *(w++) * MAP_PIXEL_SIDE_SCALE;
      /* swap coordinates */
      xi = (int)dx;
      if (xi - dx == 0.0)
	fprintf(f, "%d ", xi);
      else
	fprintf(f, "%f ", dx);

      xi = (int)dz;
      if (xi - dz == 0.0)
	fprintf(f, "%d ", xi);
      else
	fprintf(f, "%f ", dz);

      xi = (int)dy;
      if (xi - dy == 0.0)
	fprintf(f, "%d\n", xi);
      else
	fprintf(f, "%f\n", dy);
    }
  }

#ifdef NORMALS
  /* output vertex normals */
  s = map_normals;
  fprintf(f, "// Vertex normals\n");
  for (y = 0; y < p; y++) {
    for (x = 0; x < p; x++) {
      dx = *(s++);
      dy = *(s++);
      dz = *(s++);
      /* swap coordinates */
      fprintf(f, "%f %f %f\n", dx, dz, dy);
    }
  }

  /* output face normals */
  s = map_triangle_normals;
  fprintf(f, "// Face normals\n");
  for (y = 0; y < q; y++) {
    dx = *(s++);
    dy = *(s++);
    dz = *(s++);
    /* swap coordinates */
    fprintf(f, "%f %f %f\n", dx, dz, dy);
  }
#endif

  /* output triangles */
  m = map_triangles;
  r = map_area;
  ms = map_slopes;
  fprintf(f, "// Triangle data\n");
  for (a = 0; a < size; a++) {
    for (b = 0; b < size; b++) {
      /* output four triangles per run */
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[0], m[1], m[2], 0.0+b, 0.5+a, 0.5+b, 0.5+a, 0.0+b, 0.0+a);
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[3], m[4], m[5], 0.0+b, 0.0+a, 0.5+b, 0.5+a, 0.5+b, 0.0+a);
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[6], m[7], m[8], 0.5+b, 0.0+a, 0.5+b, 0.5+a, 1.0+b, 0.0+a);
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[9], m[10], m[11], 1.0+b, 0.0+a, 0.5+b, 0.5+a, 1.0+b, 0.5+a);
      m += 3*4;
      r++;
    }
    r -= size;
    for (b = 0; b < size; b++) {
      /* output four triangles per run */
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[0], m[1], m[2], 0.0+b, 1.0+a, 0.5+b, 1.0+a, 0.0+b, 0.5+a);
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[3], m[4], m[5], 0.0+b, 0.5+a, 0.5+b, 1.0+a, 0.5+b, 0.5+a);
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[6], m[7], m[8], 0.5+b, 0.5+a, 0.5+b, 1.0+a, 1.0+b, 0.5+a);
      t = (int)*r;
      if (t == MAP_AREA_GRASSLANDS) {
	if (*ms < 0.8)
	  t = 3;
	else if (map_min_max[a*size + b] == MAP_MIN)
	  t = 4;
      }
      ms++;
      fprintf(f, "%d %d %d %d %f %f %f %f %f %f\n", t, m[9], m[10], m[11], 1.0+b, 0.5+a, 0.5+b, 1.0+a, 1.0+b, 1.0+a);
      m += 3*4;
      r++;
    }
  }

  /* output textures */
  fprintf(f, "// Texture data\n");
  fprintf(f, "Textures: %d\n", 5);
  fprintf(f, "\"undersea.s3b\"\n");
  fprintf(f, "\"grasslands.s3b\"\n");
  fprintf(f, "\"mountain.s3b\"\n");
  fprintf(f, "\"drygrasslands.s3b\"\n");
  fprintf(f, "\"wetgrasslands.s3b\"\n");

  fclose(f);

  return SUCCEEDED;
}


int map_generator_save_plants(void) {

  struct plant *o;
  FILE *f;
  int i;


  f = fopen("plants.txt", "wb");
  if (f == NULL) {
    fprintf(stderr, "MAP_GENERATOR_SAVE_PLANTS: Could not open file \"plants.txt\" for writing.\n");
    return FAILED;
  }

  for (i = 0; i < size*size*8; i++) {
    o = obj_plants[i];
    while (o != NULL) {
      fprintf(f, "%f %f %f %f %f %f %d\n", o->x*MAP_PIXEL_SIDE_SCALE, o->z*MAP_PIXEL_SIDE_SCALE, o->y*MAP_PIXEL_SIDE_SCALE, o->rx, o->ry, o->rz, o->type);
      o = o->next;
    }
  }

  fclose(f);

  return SUCCEEDED;
}


int map_generator_save_psystems(void) {

  struct psystem *p;
  FILE *f;


  f = fopen("psystems.txt", "wb");
  if (f == NULL) {
    fprintf(stderr, "MAP_GENERATOR_SAVE_PLANTS: Could not open file \"psystems.txt\" for writing.\n");
    return FAILED;
  }

  p = psystems;
  while (p != NULL) {
    fprintf(f, "%d %f %f %f %f %f %f\n", p->type, p->x, p->y, p->z, p->rx, p->ry, p->rz);
    p = p->next;
  }

  fclose(f);

  return SUCCEEDED;
}
