
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>

#include "defines.h"
#include "main.h"
#include "id.h"
#include "game_object.h"
#include "player.h"
#include "action.h"
#include "psystems.h"
#include "../common/parse.h"
#include "../common/socket.h"
#include "../common/dynmem.h"

/* 1 = normal, 3 = all */
/*
#define NETWORK_DEBUG  1
#define COMMON_DEBUG   1
*/
#define USE_DEBUG_TIME 1

extern struct player *players;
extern int world_x, world_y, world_z;
extern unsigned int current_time, *world_time;
extern struct game_object **game_object_map;
extern struct psystem **psystem_map;

char *comm_buff = NULL;
int comm_buff_size = 0;

double starting_points[] = {
  61, 60, 60,
  64, 60, 60,
  66, 60, 60
};


int player_create(int s, int po, int so) {

  struct sockaddr_in clientname;
  struct player *p;
  unsigned int id;
  size_t size;
  char tmp[64];
  int i;


  p = malloc(sizeof(struct player));
  if (p == NULL) {
    fprintf(stderr, "PLAYER_CREATE: Out of memory error.\n");
    return FAILED;
  }

  p->go = malloc(sizeof(struct game_object));
  if (p->go == NULL) {
    fprintf(stderr, "PLAYER_CREATE: Out of memory error.\n");
    free(p);
    return FAILED;
  }

  p->comm_buffer_size = 64*1024;
  p->comm_buffer_i = 5;
  p->comm_buffer = malloc(p->comm_buffer_size);
  if (p->comm_buffer == NULL) {
    fprintf(stderr, "PLAYER_CREATE: Out of memory error.\n");
    free(p->go);
    free(p);
    return FAILED;
  }

  p->world_time = malloc(sizeof(unsigned int) * world_x*world_y*world_z);
  if (p->world_time == NULL) {
    fprintf(stderr, "PLAYER_CREATE: Out of memory error.\n");
    free(p->comm_buffer);
    free(p->go);
    free(p);
    return FAILED;
  }

  /* zero the cube times */
  for (i = 0; i < world_x*world_y*world_z; i++)
    p->world_time[i] = 0;

  size = sizeof(clientname);
  if (getsockname(s, (struct sockaddr *)&clientname, &size) < 0) {
    fprintf(stderr, "PLAYER_CREATE: Unknown socket.\n");
    free(p->world_time);
    free(p->comm_buffer);
    free(p->go);
    free(p);
    return FAILED;
  }

  generate_id(&p->go->id);

  i = clientname.sin_addr.s_addr;
  p->machine = i;
  p->socket = so;
  p->m1 = i & 0xff;
  p->m2 = (i>>8) & 0xff;
  p->m3 = (i>>16) & 0xff;
  p->m4 = (i>>24) & 0xff;
  p->port = po;
  p->visibility = 4;
  p->lantern_status = ON;
  p->go->type = OBJECT_PLAYER;
  p->go->object = 0;
  p->go->change_time = current_time;
  p->status = STATUS_STARTUP;
  p->output_bytes = 0;
  p->previous_time = 0;
  p->inventory = NULL;
  sprintf(p->go->name, "%d", p->go->id);

#ifdef COMMON_DEBUG
  fprintf(stderr, "PLAYER_CREATE: New player %d.%d.%d.%d:%d\n", p->m1, p->m2, p->m3, p->m4, p->port);
#endif

  player_relocate(p);
  game_object_add(p->go);

  if (players != NULL)
    players->prev = p;

  p->next = players;
  p->prev = NULL;
  players = p;

  /* send the player his position */
  tmp[0] = 'P';
  write_int(tmp + 1, p->go->id);
  i = 5;
  write_double_a(tmp, &i, p->go->x);
  write_double_a(tmp, &i, p->go->y);
  write_double_a(tmp, &i, p->go->z);
  player_write(p, tmp, i);

  /* generate dummy objects */
  for (i = 0; i < DUMMY_OBJECTS; i++) {
    if (generate_id(&id) == FAILED) {
      free(p->world_time);
      free(p->comm_buffer);
      free(p->go);
      free(p);
      return FAILED;
    }
    p->dummy_objects[i] = id;
    tmp[0] = 'd';
    write_int(tmp + 1, id);
    player_write(p, tmp, 5);
  }

  p->dummy_objects_i = DUMMY_OBJECTS;

  return SUCCEEDED;
}


void player_relocate(struct player *p) {

  int i;


  i = rand() % MAX_STARTING_POINTS;
  p->go->x = starting_points[i*3 + 0];
  p->go->y = starting_points[i*3 + 1];
  p->go->z = starting_points[i*3 + 2];
  p->go->ax = 0;
  p->go->ay = 0;
  p->go->az = 0;
}


struct player *player_get(int i) {

  struct player *p;


  p = players;
  while (p != NULL) {
    if (p->socket == i)
      return p;
    p = p->next;
  }

  fprintf(stderr, "PLAYER_GET: Player %d not found!\n", i);

  return NULL;
}


int player_change_socket(int i, int s) {

  struct player *p;


  p = player_get(i);
  if (p == NULL)
    return FAILED;

  p->socket = s;

  return SUCCEEDED;
}


int player_change_status(int i, int s) {

  struct player *p;


  p = player_get(i);
  if (p == NULL)
    return FAILED;

  p->status = s;

  return SUCCEEDED;
}


void player_exit(struct player *p) {

  struct inventory_item *i, *t;
  int n;


  fprintf(stderr, "PLAYER_EXIT: Player \"%s\" has left the game.\n", p->go->name);
  game_object_remove(p->go);

  /* free the dummy object ids */
  for (n = 0; n < p->dummy_objects_i; n++) {
    free_id(p->dummy_objects[n]);
  }

  if (p->next != NULL)
    p->next->prev = p->prev;
  if (p->prev != NULL)
    p->prev->next = p->next;
  if (p == players)
    players = p->next;

  free(p->comm_buffer);
  free(p->world_time);
  i = p->inventory;
  while (i != NULL) {
    t = i->next;
    free(i);
    i = t;
  }

  socket_delete(p->socket);
  free(p);
}


void player_carousel(void) {

  struct player *p, *t;


  /* handle player i/o */
  p = players;
  while (p != NULL) {
    if (p->status == STATUS_READY)
      player_handle(p);
    p = p->next;
  }

  /* remove quitted players */
  p = players;
  while (p != NULL) {
    if (p->status == STATUS_QUIT) {
      t = p->next;
      player_exit(p);
      p = t;
    }
    else
      p = p->next;
  }

  /* tick */
  current_time++;
}


void player_handle(struct player *p) {

  int ax, ay, az, ex, ey, ez, i, x, y, z;
  struct timeval t;
  fd_set fs;


  /* first check if the player has sent the server any data */
  FD_ZERO(&fs);
  FD_SET(p->socket, &fs);
  t.tv_sec = 0;
  t.tv_usec = 0;
  i = select(FD_SETSIZE, &fs, NULL, NULL, &t);

  /* error? */
  if (i < 0) {
    fprintf(stderr, "PLAYER_HANDLE: Select on player \"%s\" failed.\n", p->go->name);
    _exit(EXIT_FAILURE);
  }
  /* read and analyze player's data */
  if (i > 0)
    player_read_data(p);

  ax = (p->go->x / CUBE_SIDE) - p->visibility;
  ay = (p->go->y / CUBE_SIDE) - p->visibility;
  az = (p->go->z / CUBE_SIDE) - p->visibility;
  ex = ax + (p->visibility<<1) + 1;
  ey = ay + (p->visibility<<1) + 1;
  ez = az + (p->visibility<<1) + 1;

  if (ax < 0)
    ax = 0;
  if (ay < 0)
    ay = 0;
  if (az < 0)
    az = 0;
  if (ex > world_x)
    ex = world_x;
  if (ey > world_y)
    ey = world_y;
  if (ez > world_z)
    ez = world_z;

  /*
  player_send_time(p);
  */

  for (z = az; z < ez; z++) {
    for (y = ay; y < ey; y++) {
      for (x = ax; x < ex; x++) {
	i = z*world_y*world_x + y*world_x + x;
	/* has the player old data? */
	if (world_time[i] > p->world_time[i]) {
	  if (game_object_map[i] != NULL)
	    player_send_cube(p, i);
	  if (psystem_map[i] != NULL)
	    player_send_psystems(p, i);
	  p->world_time[i] = world_time[i];
	}
      }
    }
  }

  /* send player the data */
  if (p->comm_buffer_i > 5) {
    p->comm_buffer[0] = 's';
    write_int(p->comm_buffer + 1, p->comm_buffer_i - 5);
    send(p->socket, p->comm_buffer, p->comm_buffer_i, 0);

    /* player p has been handled -> reset the p's output buffer */
    p->output_bytes += p->comm_buffer_i;
#if (NETWORK_DEBUG & 2) == 2
    fprintf(stderr, "PLAYER_HANDLE: Sent %d bytes to player \"%s\" (total: %d bytes).\n", p->comm_buffer_i, p->go->name, p->output_bytes);
#endif
    p->comm_buffer_i = 5;
  }
}


void player_send_time(struct player *p) {

  struct tm *ti;
  time_t t;
  char c[16];


  time(&t);
  /* send the time only once a second */
  if (t == p->previous_time)
    return;

  ti = localtime(&t);
  c[0] = 'T';
#ifdef USE_DEBUG_TIME
  c[1] = 16;
  c[2] = 0;
  c[3] = 0;
#else
  c[1] = ti->tm_hour;
  c[2] = ti->tm_min;
  c[3] = ti->tm_sec;
#endif
  player_write(p, c, 4);
  p->previous_time = t;
}


void player_send_psystems(struct player *p, int i) {

  struct psystem *ps;
  char tmp[256];
  int x;


  ps = psystem_map[i];
  while (ps != NULL) {
    /* does the player know the object's latest status? */
    if (p->world_time[i] < ps->change_time) {
      tmp[0] = 'p';
      write_int(tmp + 1, ps->id);
      x = 5;
      write_double_a(tmp, &x, ps->x);
      write_double_a(tmp, &x, ps->y);
      write_double_a(tmp, &x, ps->z);
      write_float_a(tmp, &x, ps->rx);
      write_float_a(tmp, &x, ps->ry);
      write_float_a(tmp, &x, ps->rz);
      player_write(p, tmp, x);
    }
    ps = ps->next;
  }
}


void player_send_cube(struct player *p, int i) {

  struct game_object *g;
  char tmp[256];
  int x;


  g = game_object_map[i];
  while (g != NULL) {
    /* does the player know the object's latest status? */
    if (p->world_time[i] < g->change_time) {
      tmp[0] = 'O';
      write_int(tmp + 1, g->id);
      x = 5;
      write_double_a(tmp, &x, g->x);
      write_double_a(tmp, &x, g->y);
      write_double_a(tmp, &x, g->z);
      write_float_a(tmp, &x, g->ax);
      write_float_a(tmp, &x, g->ay);
      write_float_a(tmp, &x, g->az);
      player_write(p, tmp, x);
    }
    g = g->next;
  }
}


void player_write(struct player *p, char *d, int i) {

  int b;


  b = p->comm_buffer_i;
  while (i > 0) {
    if (b >= p->comm_buffer_size) {
      /* out of buffer -> double the size of the old buffer */
      fprintf(stderr, "PLAYER_WRITE: Doubling player \"%s\"'s output buffer to %d bytes.\n", p->go->name, p->comm_buffer_size<<1);
      if (p->comm_buffer_size == 0)
	return;
      p->comm_buffer_size *= 2;
      p->comm_buffer = realloc(p->comm_buffer, p->comm_buffer_size);
      if (p->comm_buffer == NULL) {
	fprintf(stderr, "PLAYER_WRITE: Out of memory error.\n");
	p->comm_buffer_size = 0;
	p->comm_buffer_i = 0;
	/* major error -> drop the player */
	p->status = STATUS_QUIT;
	return;
      }
    }
    p->comm_buffer[b++] = *(d++);
    i--;
  }
  p->comm_buffer_i = b;
}


void player_read_data(struct player *p) {

  char t[128], b[128], *tmp;
  struct game_object *g;
  struct psystem *ps;
  struct player *l;
  int i, x, z, size, n;


  i = recv(p->socket, t, 5, 0);

  /* get at least five bytes (to get the message size) */
  while (i < 5)
    i += recv(p->socket, t + i, 5 - i, 0);

  /* parse the size */
  if (t[0] != 's') {
    fprintf(stderr, "PLAYER_READ_DATA: Unknown command '%c', should be 's'.\n", t[0]);
    exit(0);
  }

  read_int(t + 1, &size);

  if (size > comm_buff_size) {
    fprintf(stderr, "PLAYER_READ_DATA: Data overflow from player \"%s\" (got a 5+%d byte message, room for only %d bytes).\n", p->go->name, size, comm_buff_size);
    fprintf(stderr, "PLAYER_READ_DATA: Enlarging buffers.\n");
    dynmem_handle_buffer(&comm_buff, &comm_buff_size, size);
  }

  tmp = comm_buff;
  i = 0;
  /* get the whole message (and nothing extra) */
  while (i < size)
    i += recv(p->socket, tmp + i, size - i, 0);

#if (NETWORK_DEBUG & 2) == 2
  fprintf(stderr, "PLAYER_READ_DATA: Got %d bytes from player \"%s\".\n", i, p->go->name);
#endif

  for (x = 0; x < i; ) {
#if (NETWORK_DEBUG & 2) == 2
    fprintf(stderr, "PLAYER_READ_DATA: Next command '%c' (%d).\n", tmp[x], tmp[x]);
#endif
    switch(tmp[x]) {

    /* player's visibility changed */
    case 'V':
      p->visibility = tmp[x + 1];
      x += 2;
#ifdef NETWORK_DEBUG
  fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\"'s visibility changed to %d.\n", p->go->name, p->visibility);
#endif
      continue;

    /* player wants to pick up this object */
    case 'P':
      x++;
      read_int_a(tmp, &x, &z);
#ifdef NETWORK_DEBUG
      fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" wants to pick up object %d.\n", p->go->name, z);
#endif
      g = game_object_get(z);
      if (g != NULL) {
	/* give it to the player */
	add_to_player_inventory(p, g);
	game_object_remove(g);
#ifdef NETWORK_DEBUG
	fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" picked up object %d.\n", p->go->name, z);
#endif
      }
      else {
	/* player wants an object that doesn't exist? tell him to delete it */
	b[0] = 'X';
	n = 1;
	write_int_a(b, &n, z);
	player_write(p, b, n);
      }
      continue;

    /* object station change */
    case 'O':
      x++;
      read_int_a(tmp, &x, &z);
      g = game_object_get(z);
      /* missing object? */
      if (g == NULL) {
	fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" refers to a missing object %d.\n", p->go->name, z);
	x += sizeof(double)*3 + sizeof(float)*3;
      }
      else {
	read_double_a(tmp, &x, &(g->x));
	read_double_a(tmp, &x, &(g->y));
	read_double_a(tmp, &x, &(g->z));
	read_float_a(tmp, &x, &(g->ax));
	read_float_a(tmp, &x, &(g->ay));
	read_float_a(tmp, &x, &(g->az));

	b[0] = 'O';
	write_int(b + 1, g->id);
	z = 5;
	write_double_a(b, &z, g->x);
	write_double_a(b, &z, g->y);
	write_double_a(b, &z, g->z);
	write_float_a(b, &z, g->ax);
	write_float_a(b, &z, g->ay);
	write_float_a(b, &z, g->az);
	send_to_all_but(p, b, z);
	game_object_relocate(g);
      }
      continue;

    /* a new object! */
    case 'n':
      {
	struct game_object *g;
	double px, py, pz;
	float ax, ay, az;
	unsigned int id;
	int type, object, status, k, o;


	o = x++;
	read_int_a(tmp, &x, &z);
	read_int_a(tmp, &x, &object);
	read_double_a(tmp, &x, &px);
	read_double_a(tmp, &x, &py);
	read_double_a(tmp, &x, &pz);
	read_float_a(tmp, &x, &ax);
	read_float_a(tmp, &x, &ay);
	read_float_a(tmp, &x, &az);

	/* read the name */
	for (k = 0; tmp[x] != 0; x++, k++)
	  b[k] = tmp[x];
	b[k] = 0;
	x++;

	status = tmp[x++] & GO_COLLISION;
	type = OBJECT_PICKABLE;

	for (k = 0; k < DUMMY_OBJECTS; k++) {
	  if (p->dummy_objects[k] == z) {
	    break;
	  }
	}

	/* no match!? */
	if (k == DUMMY_OBJECTS) {
	  fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" uses id %d, which it doesn't own.\n", p->go->name, z);
	  continue;
	}

	game_object_create(b, z, type, object, px, py, pz, ax, ay, az, &g, status);

	/* let the others know about this object */
	tmp[o] = 'N';
	send_to_all_but(p, tmp+o, x-o);

	p->dummy_objects_i--;

	/* generate a new dummy id */
	if (generate_id(&id) == FAILED)
	  continue;

	p->dummy_objects[k] = id;
	p->dummy_objects_i++;

	/* send him back a new dummy object/id */
	b[0] = 'd';
	write_int(b + 1, id);
	player_write(p, b, 5);

	continue;
      }

    /* object request */
    case 'M':
      x++;
      read_int_a(tmp, &x, &z);
      g = game_object_get(z);
      /* missing object? */
      if (g != NULL) {
	b[0] = 'N';
	z = 1;
	write_int_a(b, &z, g->id);
	write_int_a(b, &z, g->object);
	write_double_a(b, &z, g->x);
	write_double_a(b, &z, g->y);
	write_double_a(b, &z, g->z);
	write_float_a(b, &z, g->ax);
	write_float_a(b, &z, g->ay);
	write_float_a(b, &z, g->az);
	sprintf(b + z, "%s%c", g->name, 0);
	z += strlen(g->name) + 1;
	b[z++] = g->status;
	player_write(p, b, z);
      }
      continue;

    /* particle system request */
    case 'm':
      x++;
      read_int_a(tmp, &x, &z);
      ps = psystems_get(z);

      if (ps != NULL) {
	b[0] = 'n';
	z = 1;
	write_int_a(b, &z, ps->id);
	write_int_a(b, &z, ps->type);
	write_double_a(b, &z, ps->x);
	write_double_a(b, &z, ps->y);
	write_double_a(b, &z, ps->z);
	write_float_a(b, &z, ps->rx);
	write_float_a(b, &z, ps->ry);
	write_float_a(b, &z, ps->rz);
	player_write(p, b, z);
      }
      continue;

    /* name change */
    case 'N':
      strcpy(b, tmp + x + 1);
      x += strlen(b) + 2;
#ifdef NETWORK_DEBUG
      fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" wants to change his name to \"%s\".\n", p->go->name, b);
#endif

      /* is the name in use already? */
      l = players;
      while (l != NULL) {
	if (l->status == STATUS_READY && strcmp(l->go->name, b) == 0)
	  break;
	l = l->next;
      }

      /* name change accepted */
      if (l == NULL) {
	fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" changed his name to \"%s\".\n", p->go->name, b);
	strcpy(p->go->name, b);
	b[0] = 'Y';
	write_int(b + 1, p->go->id);
	sprintf(b + 5, "%s%c", p->go->name, 0);
	send_to_all(b, strlen(p->go->name) + 6);
      }
      continue;

    /* lantern status change */
    case 'L':
      x++;
      p->lantern_status = tmp[x++];
#ifdef NETWORK_DEBUG
      fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\"'s new lantern status = %d.\n", p->go->name, p->lantern_status);
#endif
      continue;

    /* client exit */
    case 'D':
      x++;
      p->status = STATUS_QUIT;
      b[0] = 'D';
      player_write(p, b, 1);
#ifdef NETWORK_DEBUG
      fprintf(stderr, "PLAYER_READ_DATA: Player \"%s\" has quitted.\n", p->go->name);
#endif
      continue;

    default:
      fprintf(stderr, "PLAYER_READ_DATA: Unknown command '%d'.\n", tmp[x]);
      x++;
      continue;
    }
  }
}
